from __future__ import annotations

from typing import Callable

import numpy as np
from numba import guvectorize
from pywt import downcoef

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def discrete_wavelet_transform(wave_type: str, level: int) -> Callable:
    """
    Apply a discrete wavelet transform to the waveform and return only
    the approximate coefficients.

    Note
    ----
    This processor is composed of a factory function that is called using the
    ``init_args`` argument. The input and output waveforms are passed using
    ``args``. The output waveform dimension must be specified.


    Parameters
    ----------
    wave_type
       The wavelet type for discrete convolution ``('haar', 'db1', ...)``
    level
       The level of decompositions to be performed ``(1, 2, ...)``


    JSON Configuration Example
    --------------------------
    .. code-block :: json

        "dwt":{
            "function": "discrete_wavelet_transform",
            "module": "pygama.dsp.processors",
            "args": ["wf_blsub", "dwt(250)"],
            "unit": "ADC",
            "prereqs": ["wf_blsub"],
            "init_args": ["'haar'", "3",]
        }
    """

    if level <= 0:
        raise DSPFatal("The level must be a positive integer")

    @guvectorize(
        ["void(float32[:], float32[:])", "void(float64[:], float64[:])"],
        "(n),(m)",
        **nb_kwargs,
        forceobj=True,
    )
    def dwt_out(w_in: np.ndarray, w_out: np.ndarray) -> None:
        """
        Parameters
        ----------
        w_in
           The input waveform
        w_out
           The approximate coefficients. The dimension of this array can be calculated
           by ``out_dim = wf_length/(filter_length^level)``, where ``filter_length``
           can be obtained via ``pywt.Wavelet(wave_type).dec_len``.
        """
        w_out[:] = np.nan

        if np.isnan(w_in).any():
            return

        w_out[:] = downcoef("a", w_in, wave_type, level=level)

    return dwt_out



import numpy as np
import pytest
from pytest import approx

from pygama import lgdo
from pygama.raw.fc.fc_status_decoder import FCStatusDecoder
from pygama.raw.raw_buffer import RawBuffer


@pytest.fixture(scope="module")
def status_rb(fcio_obj):
    decoder = FCStatusDecoder()
    rb = RawBuffer(lgdo=decoder.make_lgdo(size=1))

    # just get the first status record in the file and exit
    while True:
        pid = fcio_obj.get_record()
        assert pid > 0  # make sure there is at least one status record in the file
        if pid == 4:
            decoder.decode_packet(fcio=fcio_obj, status_rb=rb, packet_id=420)
            break

    return rb


def test_decoding(status_rb):
    assert status_rb.is_full() is True


def test_data_types(status_rb):
    tbl = status_rb.lgdo

    assert isinstance(tbl["status"], lgdo.Array)
    assert isinstance(tbl["statustime"], lgdo.Array)
    assert isinstance(tbl["cputime"], lgdo.Array)
    assert isinstance(tbl["startoffset"], lgdo.Array)
    assert isinstance(tbl["cards"], lgdo.Array)
    assert isinstance(tbl["size"], lgdo.Array)
    assert isinstance(tbl["environment"], lgdo.ArrayOfEqualSizedArrays)
    assert isinstance(tbl["totalerrors"], lgdo.Array)
    assert isinstance(tbl["linkerrors"], lgdo.Array)
    assert isinstance(tbl["ctierrors"], lgdo.Array)
    assert isinstance(tbl["enverrors"], lgdo.Array)
    assert isinstance(tbl["othererrors"], lgdo.ArrayOfEqualSizedArrays)


def test_values(status_rb, fcio_obj):
    fc = fcio_obj
    tbl = status_rb.lgdo
    i = status_rb.loc - 1

    assert i == 0
    assert tbl["status"].nda[i] == fc.status
    assert tbl["statustime"].nda[i] == approx(fc.statustime[0] + fc.statustime[1] / 1e6)
    assert tbl["cputime"].nda[i] == approx(fc.statustime[2] + fc.statustime[3] / 1e6)
    assert tbl["startoffset"].nda[i] == approx(
        fc.statustime[5] + fc.statustime[6] / 1e6
    )
    assert tbl["cards"].nda[i] == fc.cards
    assert tbl["size"].nda[i] == fc.size
    assert np.array_equal(tbl["environment"].nda[i], fc.environment)
    assert tbl["totalerrors"].nda[i] == fc.totalerrors
    assert tbl["linkerrors"].nda[i] == fc.linkerrors
    assert tbl["ctierrors"].nda[i] == fc.ctierrors
    assert tbl["enverrors"].nda[i] == fc.enverrors
    assert np.array_equal(tbl["othererrors"].nda[i], fc.othererrors)



import json
from pathlib import Path

import pytest

from pygama.flow import FileDB

config_dir = Path(__file__).parent / "configs"


@pytest.fixture(scope="module")
def test_filedb(lgnd_test_data):
    with open(config_dir / "filedb-config.json") as f:
        config = json.load(f)

    config["data_dir"] = lgnd_test_data.get_path("lh5/prod-ref-l200/generated/tier")
    return FileDB(config)


@pytest.fixture(scope="module")
def test_filedb_full(lgnd_test_data):
    with open(config_dir / "filedb-config.json") as f:
        config = json.load(f)

    config["data_dir"] = lgnd_test_data.get_path("lh5/prod-ref-l200/generated/tier")
    db = FileDB(config)
    db.scan_tables_columns()
    return db



from __future__ import annotations

import fcutils

from pygama.raw.data_decoder import DataDecoder, RawBuffer


class FCStatusDecoder(DataDecoder):
    """Decode FlashCam digitizer status data."""

    def __init__(self, *args, **kwargs) -> None:
        self.decoded_values = {
            # 0: Errors occurred, 1: no errors
            "status": {"dtype": "int32"},
            # fc250 seconds, microseconds, dummy, startsec startusec
            "statustime": {"dtype": "float32", "units": "s"},
            # CPU seconds, microseconds, dummy, startsec startusec
            "cputime": {"dtype": "float64", "units": "s"},
            # fc250 seconds, microseconds, dummy, startsec startusec
            "startoffset": {"dtype": "float32", "units": "s"},
            # Total number of cards (number of status data to follow)
            "cards": {"dtype": "int32"},
            # Size of each status data
            "size": {"dtype": "int32"},
            # FC card-wise environment status
            "environment": {
                # Array contents:
                # [0-4] Temps in mDeg
                # [5-10] Voltages in mV
                # 11 main current in mA
                # 12 humidity in o/oo
                # [13-14] Temps from adc cards in mDeg
                # FIXME: change to a table?
                "dtype": "uint32",
                "datatype": "array_of_equalsized_arrays<1,1>{real}",
                "length": 16,
            },
            # FC card-wise list DAQ errors during data taking
            "totalerrors": {"dtype": "uint32"},
            "enverrors": {"dtype": "uint32"},
            "ctierrors": {"dtype": "uint32"},
            "linkerrors": {"dtype": "uint32"},
            "othererrors": {
                "dtype": "uint32",
                "datatype": "array_of_equalsized_arrays<1,1>{real}",
                "length": 5,
            },
        }
        """Default FlashCam status decoded values.

        Warning
        -------
        This configuration can be dynamically modified by the decoder at runtime.
        """
        super().__init__(*args, **kwargs)

    def decode_packet(
        self, fcio: fcutils.fcio, status_rb: RawBuffer, packet_id: int
    ) -> bool:
        # aliases for brevity
        tbl = status_rb.lgdo
        ii = status_rb.loc

        # status -- 0: Errors occurred, 1: no errors
        tbl["status"].nda[ii] = fcio.status

        # times
        tbl["statustime"].nda[ii] = fcio.statustime[0] + fcio.statustime[1] / 1e6
        tbl["cputime"].nda[ii] = fcio.statustime[2] + fcio.statustime[3] / 1e6
        tbl["startoffset"].nda[ii] = fcio.statustime[5] + fcio.statustime[6] / 1e6

        # Total number of cards (number of status data to follow)
        tbl["cards"].nda[ii] = fcio.cards

        # Size of each status data
        tbl["size"].nda[ii] = fcio.size

        # FC card-wise environment status (temp., volt., hum., ...)
        tbl["environment"].nda[ii][:] = fcio.environment

        # FC card-wise list DAQ errors during data taking
        tbl["totalerrors"].nda[ii] = fcio.totalerrors
        tbl["linkerrors"].nda[ii] = fcio.linkerrors
        tbl["ctierrors"].nda[ii] = fcio.ctierrors
        tbl["enverrors"].nda[ii] = fcio.enverrors
        tbl["othererrors"].nda[ii][:] = fcio.othererrors

        status_rb.loc += 1

        return status_rb.loc >= len(tbl)



import pytest

from pygama import lgdo
from pygama.raw.orca.orca_run_decoder import ORRunDecoderForRun
from pygama.raw.raw_buffer import RawBuffer


@pytest.fixture(scope="module")
def run_rbkd(orca_stream):
    decoder = ORRunDecoderForRun(header=orca_stream.header)

    rbkd = {}
    rbkd[0] = RawBuffer(lgdo=decoder.make_lgdo(size=1))

    this_packet = orca_stream.load_packet(skip_unknown_ids=False)
    orca_stream.close_stream()  # avoid leaving file open

    # assert correct type for ORRunDecoderForRun
    assert (this_packet[0] >> 18) == orca_stream.header["dataDescription"][
        "ORRunModel"
    ]["Run"]["dataId"] >> 18

    # assert that it worked and the buffer is full
    assert decoder.decode_packet(packet=this_packet, packet_id=1, rbl=rbkd)

    return rbkd, this_packet


def test_decoding(run_rbkd):
    assert run_rbkd[0] != {}


def test_data_types(run_rbkd):
    for _, v in run_rbkd[0].items():
        tbl = v.lgdo
        assert isinstance(tbl, lgdo.Struct)
        assert isinstance(tbl["subrun_number"], lgdo.Array)
        assert isinstance(tbl["runstartorstop"], lgdo.Array)
        assert isinstance(tbl["quickstartrun"], lgdo.Array)
        assert isinstance(tbl["remotecontrolrun"], lgdo.Array)
        assert isinstance(tbl["heartbeatrecord"], lgdo.Array)
        assert isinstance(tbl["endsubrunrecord"], lgdo.Array)
        assert isinstance(tbl["startsubrunrecord"], lgdo.Array)
        assert isinstance(tbl["run_number"], lgdo.Array)
        assert isinstance(tbl["time"], lgdo.Array)


def test_values(run_rbkd):
    this_packet = run_rbkd[1]

    decoded_values = ORRunDecoderForRun().decoded_values

    for _, v in run_rbkd[0].items():
        loc = v.loc - 1
        tbl = v.lgdo

        assert tbl["subrun_number"].nda[loc] == (this_packet[1] & 0xFFFF0000) >> 16

        for i, k in enumerate(decoded_values):
            if 0 < i < 7:
                assert tbl[k].nda[loc] == (this_packet[1] & (1 << (i - 1))) >> (i - 1)

        assert tbl["run_number"].nda[loc] == this_packet[2]
        assert tbl["time"].nda[loc] == this_packet[3]



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:], float32[:], float32[:])",
        "void(float64[:], float64[:], float64[:], float64[:], float64[:])",
    ],
    "(n)->(),(),(),()",
    **nb_kwargs,
)
def linear_slope_fit(
    w_in: np.ndarray, mean: float, stdev: float, slope: float, intercept: float
) -> None:
    """
    Calculate the mean and standard deviation of the waveform using
    Welford's method as well as the slope an intercept of the waveform
    using linear regression.

    Parameters
    ----------
    w_in
        the input waveform.
    mean
        the mean of the waveform.
    stdev
        the standard deviation of the waveform.
    slope
        the slope of the linear fit.
    intercept
        the intercept of the linear fit.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "bl_mean, bl_std, bl_slope, bl_intercept": {
            "function": "linear_slope_fit",
            "module": "pygama.dsp.processors",
            "args": ["wf_blsub[0:1650]", "bl_mean", "bl_std", "bl_slope", "bl_intercept"],
            "unit": ["ADC", "ADC", "ADC", "ADC"],
        }
    """
    mean[0] = np.nan
    stdev[0] = np.nan
    slope[0] = np.nan
    intercept[0] = np.nan

    if np.isnan(w_in).any():
        return

    sum_x = sum_x2 = sum_xy = sum_y = mean[0] = stdev[0] = 0
    isum = len(w_in)

    for i in range(0, len(w_in), 1):
        # the mean and standard deviation
        temp = w_in[i] - mean
        mean += temp / (i + 1)
        stdev += temp * (w_in[i] - mean)

        # linear regression
        sum_x += i
        sum_x2 += i * i
        sum_xy += w_in[i] * i
        sum_y += w_in[i]

    stdev /= isum - 1
    np.sqrt(stdev, stdev)

    slope[0] = (isum * sum_xy - sum_x * sum_y) / (isum * sum_x2 - sum_x * sum_x)
    intercept[0] = (sum_y - sum_x * slope[0]) / isum


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:], float32[:], float32[:])",
        "void(float64[:], float64[:], float64[:], float64[:], float64[:])",
    ],
    "(n),(),()->(),()",
    **nb_kwargs,
)
def linear_slope_diff(
    w_in: np.ndarray, slope: float, intercept: float, mean: float, rms: float
) -> None:
    """
    Calculate the mean, standard deviation, slope and y-intercept of waveform.

    Uses Welford's method and linear regression.

    Parameters
    ----------
    w_in
        the input waveform.
    mean
        the mean of the waveform.
    stdev
        the standard deviation of the waveform.
    slope
        the slope of the linear fit.
    intercept
        the intercept of the linear fit.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "bl_mean, bl_std, bl_slope, bl_intercept": {
            "function": "linear_slope_fit",
            "module": "pygama.dsp.processors",
            "args": ["wf_blsub[0:1650]", "bl_mean", "bl_std", "bl_slope", "bl_intercept"],
            "unit": ["ADC", "ADC", "ADC", "ADC"],
        }
    """
    mean[0] = np.nan
    rms[0] = np.nan

    if np.isnan(w_in).any() or np.isnan(slope) or np.isnan(intercept):
        return

    mean[0] = rms[0] = 0
    isum = len(w_in)

    for i in range(0, len(w_in), 1):
        # the mean and standard deviation
        temp = w_in[i] - (slope * i + intercept)
        mean += temp / (i + 1)
        rms += temp * temp

    rms /= isum - 1
    np.sqrt(rms, rms)



from pathlib import Path

import pandas as pd
import pytest

from pygama import lgdo
from pygama.flow import DataLoader

config_dir = Path(__file__).parent / "configs"


@pytest.fixture(scope="function")
def test_dl(test_filedb):
    return DataLoader(f"{config_dir}/data-loader-config.json", test_filedb)


def test_init(test_dl):
    pass


def test_init_variants(test_filedb):
    assert DataLoader(str(config_dir), test_filedb).config is not None
    assert (
        DataLoader(
            f"{config_dir}/nested/data-loader-config-nested.json[nest1/nest2]",
            test_filedb,
        ).config
        is not None
    )
    assert (
        DataLoader(f"{config_dir}/nested[nest1/nest2]", test_filedb).config is not None
    )


def test_simple_load(test_dl):
    test_dl.set_files("all")
    test_dl.set_output(columns=["timestamp"])
    data = test_dl.load()

    assert isinstance(data, lgdo.Table)
    assert list(data.keys()) == ["hit_table", "hit_idx", "file", "timestamp"]


def test_simple_chunked_load(test_dl):
    test_dl.set_files("all")
    test_dl.set_output(columns=["timestamp"])
    for data in test_dl.next(chunk_size=2):
        assert len(data) == 2
        assert isinstance(data, lgdo.Table)
        assert list(data.keys()) == ["hit_table", "hit_idx", "file", "timestamp"]


def test_load_wfs(test_dl):
    test_dl.set_files("all")
    test_dl.set_output(columns=["waveform"], fmt="lgdo.Table")
    data = test_dl.load()
    assert isinstance(data, lgdo.Table)
    assert list(data.keys()) == ["hit_table", "hit_idx", "file", "waveform"]

    test_dl.set_output(columns=["waveform"], fmt="pd.DataFrame")
    data = test_dl.load()
    assert isinstance(data, pd.DataFrame)
    assert list(data.keys()) == [
        "hit_table",
        "hit_idx",
        "file",
        "waveform_t0",
        "waveform_dt",
        "waveform_values",
    ]


def test_no_merge(test_dl):
    test_dl.set_files("all")
    test_dl.set_output(columns=["timestamp"], merge_files=False)
    data = test_dl.load()

    assert isinstance(data, dict)
    assert isinstance(data[0], lgdo.Table)
    assert len(data) == 2
    assert list(data[0].keys()) == ["hit_table", "hit_idx", "timestamp"]


def test_outputs(test_dl):
    test_dl.set_files("all")
    test_dl.set_output(
        fmt="pd.DataFrame", columns=["timestamp", "channel", "bl_mean", "hit_par1"]
    )
    data = test_dl.load()

    assert isinstance(data, pd.DataFrame)
    assert list(data.keys()) == [
        "hit_table",
        "hit_idx",
        "file",
        "timestamp",
        "channel",
        "bl_mean",
        "hit_par1",
    ]


def test_any_mode(test_dl):
    test_dl.filedb.scan_tables_columns()
    test_dl.set_files("all")
    test_dl.set_cuts({"hit": "daqenergy == 634"})
    el = test_dl.build_entry_list(tcm_level="tcm", mode="any")

    assert len(el) == 42


def test_set_files(test_dl):
    test_dl.set_files("timestamp == '20220716T104550Z'")
    test_dl.set_output(columns=["timestamp"], merge_files=False)
    data = test_dl.load()

    assert len(data) == 1


def test_set_keylist(test_dl):
    test_dl.set_files(["20220716T104550Z", "20220716T104550Z"])
    test_dl.set_output(columns=["timestamp"], merge_files=False)
    data = test_dl.load()

    assert len(data) == 1


def test_set_datastreams(test_dl):
    test_dl.set_files("all")
    test_dl.set_datastreams([1, 3, 8], "ch")
    test_dl.set_output(columns=["channel"], merge_files=False)
    data = test_dl.load()

    assert (data[0]["hit_table"].nda == [1, 3, 8]).all()
    assert (data[0]["channel"].nda == [1, 3, 8]).all()


def test_set_cuts(test_dl):
    test_dl.set_files("all")
    test_dl.set_cuts({"hit": "card == 3"})
    test_dl.set_output(columns=["card"])
    data = test_dl.load()

    assert (data["hit_table"].nda == [12, 13, 12, 13]).all()


def test_browse(test_dl):
    test_dl.set_files("all")
    test_dl.set_output(
        fmt="pd.DataFrame", columns=["timestamp", "channel", "bl_mean", "hit_par1"]
    )
    wb = test_dl.browse()

    wb.draw_next()



from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING

import numpy as np

import pygama.lgdo as lgdo
from pygama.dsp.errors import ProcessingChainError
from pygama.dsp.processing_chain import build_processing_chain as bpc
from pygama.lgdo import Array, ArrayOfEqualSizedArrays, Table
from pygama.lgdo.compression import WaveformCodec
from pygama.lgdo.compression.utils import str2wfcodec

if TYPE_CHECKING:
    from pygama.raw.raw_buffer import RawBuffer

log = logging.getLogger(__name__)


def buffer_processor(rb: RawBuffer) -> Table:
    r"""Process raw data buffers.

    Takes in a :class:`.RawBuffer`, performs any processes specified in the
    :class:`.RawBuffer`'s ``proc_spec`` attribute (a dictionary), and returns a
    :class:`.Table` with the processed buffer. This `tmp_table` shares columns
    with the :class:`.RawBuffer`'s LGDO (`rb.lgdo`), so no data is copied.

    Currently implemented ``proc_spec`` processors:

    ``"window": ["waveform", start_index, stop_index, "out_name"]`` `(list)`
      Windows objects with a name specified by the first argument, the window
      start and stop indices are the next two arguments, and then updates the
      `rb.lgdo` with a name specified by the last argument. If the object is an
      :class:`.WaveformTable`, then the ``t0`` attribute is updated
      accordingly.  Although it is possible to use the DSP config to perform
      windowing, this hard-coded version avoids a conversion to ``float32``.

    ``"dsp_config": { <dsp_config> }`` `(dict)`
      Performs DSP given by the ``<dsp_config>`` specification. See
      :func:`~.dsp.processing_chain.build_processing_chain` for more information
      on DSP configuration dictionaries.  All fields in the output of the DSP
      are written to the `rb.lgdo`.

    ``"drop": ["waveform" [, ...]]`` `(list)`
      Drops any requested fields from the `rb.lgdo`.

    ``"dtype_conv": {"lgdo": "dtype" [, ...]}`` `(dict)`
      Casts `lgdo` to the requested data type.

    ``"compression": { "lgdo": "codec_name" [, ...]}`` `(dict)`
      Updates the `compression` attribute of `lgdo` to `codec_name`. The
      attribute sets the compression algorithm applied by
      :func:`~.lgdo.lh5_store.LH5Store.read_object` before writing `lgdo` to
      disk.

    Parameters
    ----------
    rb
        A :class:`.RawBuffer` to be processed, must contain a `proc_spec` attribute.

    Notes
    -----
    The original ``waveforms`` column in the table is not written to file if
    request! All updates are done on the `tmp_table`, which shares the fields
    with `rb.lgdo` and are done in place. The `tmp_table` is necessary so that
    the `rb.lgdo` keeps arrays needed by the table in the buffer.  An example
    `proc_spec` in an :func:`~.raw.build_raw.build_raw` `out_spec` is below. ::

        {
          "FCEventDecoder" : {
            "g{key:0>3d}" : {
              "key_list" : [[24, 64]],
              "out_stream" : "$DATADIR/{file_key}_geds.lh5:/geds",
              "proc_spec": {
                "window": ["waveform", 100, -100, "windowed_waveform"],
                "dsp_config": {
                  "outputs": ["presummed_waveform", "t_sat_lo", "t_sat_hi"],
                  "processors": {
                    "presummed_waveform": {
                      "function": "presum",
                      "module": "pygama.dsp.processors",
                      "args": [
                        "waveform",
                        "presummed_waveform(shape=len(waveform)//16, period=waveform.period*16, offset=waveform.offset, 'f')"
                      ],
                      "unit": "ADC"
                    },
                    "t_sat_lo, t_sat_hi": {
                      "function": "saturation",
                      "module": "pygama.dsp.processors",
                      "args": ["waveform", 16, "t_sat_lo", "t_sat_hi"],
                      "unit": "ADC"
                    }
                  }
                },
                "drop": ["waveform", "packet_ids"],
                "dtype_conv": {
                  "presummed_waveform/values": "uint32",
                  "t_sat_lo": "uint16",
                  "t_sat_hi": "uint16",
                ,}
                "compression": {
                  "windowed_waveform/values": RadwareSigcompress(codec_shift=-32768),
                  "presummed_waveform/values": ULEB128ZigZagDiff(),
                }
              }
            },
            "spms" : {
              "key_list" : [ [6,23] ],
              "out_stream" : "$DATADIR/{file_key}_spms.lh5:/spms"
            }
          }
        }
    """
    # Check that there is a valid object to process
    if isinstance(rb.lgdo, lgdo.Table) or isinstance(rb.lgdo, lgdo.Struct):
        # Create the temporary table that will be written to file
        rb_table_size = rb.lgdo.size
        tmp_table = Table(size=rb_table_size)
        tmp_table.join(other_table=rb.lgdo)

    # This is needed if there is a "*" key expansion for decoders in build_raw
    # In the worst case, just return an unprocessed rb.lgdo
    else:
        log.info(f"Cannot process buffer with an lgdo of type {type(rb.lgdo)}")
        tmp_table = rb.lgdo
        return tmp_table

    # Perform windowing, if requested
    if "window" in rb.proc_spec.keys():
        process_window(rb, tmp_table)

    # Read in and perform the DSP routine
    if "dsp_config" in rb.proc_spec.keys():
        process_dsp(rb, tmp_table)

    # Cast as requested dtype before writing to the table
    if "dtype_conv" in rb.proc_spec.keys():
        process_dtype_conv(rb, tmp_table)

    # Drop any requested columns from the table
    if "drop" in rb.proc_spec.keys():
        process_drop(rb, tmp_table)

    # at last, assign compression attributes
    if "compression" in rb.proc_spec.keys():
        for name, codec in rb.proc_spec["compression"].items():
            ptr = tmp_table
            for word in name.split("/"):
                ptr = ptr[word]

            ptr.attrs["compression"] = (
                codec if isinstance(codec, WaveformCodec) else str2wfcodec(codec)
            )

    return tmp_table


def process_window(rb: RawBuffer, tmp_table: Table) -> None:
    r"""Window :class:`.ArrayOfEqualSizedArrays`.

    Windows arrays of equal sized arrays according to specifications given in
    the `rb.proc_spec` ``window`` key.

    First checks if the `rb.lgdo` is a :class:`.Table` or not. If it's not a
    table, then we only process it if its `rb.out_name` is the same as the
    window ``in_name``.

    If `rb.lgdo` is a table, special processing is done if the window
    ``in_name`` field is an :class:`.WaveformTable` in order to update the
    ``t0``\ s.  Otherwise, windowing of the field is performed without updating
    any of the other attributes.

    Parameters
    ----------
    rb
        a :class:`.RawBuffer` to be processed.
    tmp_table
        a :class:`.Table` that shares columns with the `rb.lgdo`.

    Notes
    -----
    This windowing hard-coded; it is done without calling :mod:`.dsp.build_dsp`
    to avoid a conversion to ``float32``.

    """
    # Read the window parameters from the proc_spec
    window_in_name = rb.proc_spec["window"][0]
    window_start_idx = int(rb.proc_spec["window"][1])
    window_end_idx = int(rb.proc_spec["window"][2])
    window_out_name = rb.proc_spec["window"][3]

    # Check if rb.lgdo is a table and if the window_in_name is a key
    if (isinstance(rb.lgdo, lgdo.Table) or isinstance(rb.lgdo, lgdo.Struct)) and (
        window_in_name in rb.lgdo.keys()
    ):
        # Now check if the window_in_name is a waveform table or not, if so we need to modify the t0s
        if isinstance(rb.lgdo[window_in_name], lgdo.WaveformTable):
            # modify the t0s
            t0s = process_windowed_t0(
                rb.lgdo[window_in_name].t0, rb.lgdo[window_in_name].dt, window_start_idx
            )

            # Window the waveform values
            array_of_arrays = tmp_table[window_in_name].values
            windowed_array_of_arrays = window_array_of_arrays(
                array_of_arrays, window_start_idx, window_end_idx
            )

            # Write to waveform table and then to file
            wf_table = lgdo.WaveformTable(
                t0=t0s, dt=rb.lgdo[window_in_name].dt, values=windowed_array_of_arrays
            )

            # add this wf_table to the temporary table
            tmp_table.add_field(window_out_name, wf_table, use_obj_size=True)

        # otherwise, it's (hopefully) just an array of equal sized arrays
        else:
            array_of_arrays = tmp_table[window_in_name]
            windowed_array_of_arrays = window_array_of_arrays(
                array_of_arrays, window_start_idx, window_end_idx
            )
            tmp_table.add_field(
                window_out_name, windowed_array_of_arrays, use_obj_size=True
            )

        return None

    # otherwise, rb.lgdo is some other type and we only process it if the rb.out_name is the same as window_in_name
    elif rb.out_name == window_in_name:
        array_of_arrays = tmp_table
        windowed_array_of_arrays = window_array_of_arrays(
            array_of_arrays, window_start_idx, window_end_idx
        )

        rb.out_name = window_out_name
        tmp_table = windowed_array_of_arrays

        return None

    else:
        log.info(f"{window_in_name} not a valid key for this RawBuffer")
        return None


def window_array_of_arrays(
    array_of_arrays: ArrayOfEqualSizedArrays, window_start_idx: int, window_end_idx: int
) -> ArrayOfEqualSizedArrays:
    """Given an array of equal sized arrays, for each array it returns the view
    ``[window_start_idx:window_end_idx]``.
    """
    if isinstance(array_of_arrays, lgdo.ArrayOfEqualSizedArrays):
        return array_of_arrays.nda[:, window_start_idx:window_end_idx]
    else:
        raise TypeError(
            f"Do not know how to window an LGDO of type {type(array_of_arrays)}"
        )


def process_windowed_t0(t0s: Array, dts: Array, start_index: int) -> Array:
    """In order for the processed data to work well with :mod:`.dsp.build_dsp`, we need
    to keep ``t0`` in its original units.

    So we transform ``start_index`` to the units of ``t0`` and add it to every
    ``t0`` value.
    """
    # don't want to modify the original lgdo_table t0s
    # deepcopy also preserves attributes
    copy_t0s = copy.deepcopy(t0s)

    # perform t0+start_index*dt to rewrite the new t0 in terms of sample
    start_index *= dts.nda
    copy_t0s.nda += start_index
    return copy_t0s


def process_dsp(rb: RawBuffer, tmp_table: Table) -> None:
    r"""Run a DSP processing chain.

    Run a provided DSP config from `rb.proc_spec` using
    :func:`.dsp.build_processing_chain`, and add specified outputs to the
    `rb.lgdo`.

    Parameters
    ----------
    rb
        a :class:`.RawBuffer` that contains a `proc_spec` and an `lgdo`
        attribute.
    tmp_table
        a :class:`pygama.lgdo.Table` that is temporarily created to be written
        to the raw file.

    Notes
    -----
    `rb.lgdo` is assumed to be an :class`.Table` so that multiple DSP processor
    outputs can be written to it.
    """
    # Load the dsp_dict
    dsp_dict = rb.proc_spec["dsp_config"]

    # Try building the processing chain
    try:
        # execute the processing chain
        # This checks that the rb.lgdo is a table and that the field_name is present in the table
        proc_chain, mask, dsp_out = bpc(rb.lgdo, dsp_dict)
    # Allow for exceptions, in the case of "*" key expansion in the build_raw out_spec
    except ProcessingChainError:
        log.info("DSP could not be performed")
        return None

    proc_chain.execute()

    # For every processor in dsp_dict for this group, create a new entry in the
    # lgdo table with that processor's name.  If the processor returns a
    # waveform, create a new waveform table and add it to the original lgdo
    # table
    for proc in dsp_out.keys():
        # # Check what DSP routine the processors output is from, and manipulate accordingly
        tmp_table.add_field(proc, dsp_out[proc], use_obj_size=True)

    return None


def process_dtype_conv(rb: RawBuffer, tmp_table: Table) -> None:
    """Change the types of fields in an `rb.lgdo` according to the values
    specified in the ``proc_spec``'s ``dtype_conv`` list.  It operates in place
    on `tmp_table`.

    Notes
    -----
    This assumes that name provided points to an object in the `rb.lgdo` that has
    an `nda` attribute.
    """
    type_list = rb.proc_spec["dtype_conv"]
    for return_name in type_list.keys():
        # Take care of nested tables with a for loop
        path = return_name.split("/")
        return_value = tmp_table
        for key in path:
            try:
                return_value = return_value[key]
            # Allow for exceptions in the case of "*" expansion in the build_raw out_spec
            except KeyError:
                log.info(f"{key} is not a valid key in the rb.lgdo")
                return None

        # If we have a numpy array as part of the lgdo, recast its type
        if hasattr(return_value, "nda"):
            return_value.nda = return_value.nda.astype(np.dtype(type_list[return_name]))
        else:
            raise TypeError(f"Cannot recast an object of type {type(return_value)}")

    return None


def process_drop(rb: RawBuffer, tmp_table: Table) -> None:
    """Drops any requested fields from the `rb.lgdo`."""
    for drop_keys in rb.proc_spec["drop"]:
        try:
            tmp_table.remove_column(drop_keys, delete=False)
        except KeyError:
            log.info(f"Cannot remove field {drop_keys} from rb.lgdo")
            return None



import inspect

import numpy as np
import pytest

from pygama.dsp.errors import DSPFatal
from pygama.dsp.processors import fixed_time_pickoff


def test_fixed_time_pickoff(compare_numba_vs_python):
    """Testing function for the fixed_time_pickoff processor."""

    len_wf = 20

    # test for nan if w_in has a nan
    w_in = np.ones(len_wf)
    w_in[4] = np.nan
    assert np.isnan(compare_numba_vs_python(fixed_time_pickoff, w_in, 1, ord("i")))

    # test for nan if nan is passed to t_in
    w_in = np.ones(len_wf)
    assert np.isnan(compare_numba_vs_python(fixed_time_pickoff, w_in, np.nan, ord("i")))

    # test for nan if t_in is negative
    w_in = np.ones(len_wf)
    assert np.isnan(compare_numba_vs_python(fixed_time_pickoff, w_in, -1, ord("i")))

    # test for nan if t_in is too large
    w_in = np.ones(len_wf)
    assert np.isnan(compare_numba_vs_python(fixed_time_pickoff, w_in, len_wf, ord("i")))

    # test for DSPFatal errors being raised
    # noninteger t_in with integer interpolation
    with pytest.raises(DSPFatal):
        w_in = np.ones(len_wf)
        fixed_time_pickoff(w_in, 1.5, ord("i"))
    with pytest.raises(DSPFatal):
        a_out = np.empty(len_wf)
        inspect.unwrap(fixed_time_pickoff)(w_in, 1.5, ord("i"), a_out)

    # unsupported mode_in character
    with pytest.raises(DSPFatal):
        w_in = np.ones(len_wf)
        fixed_time_pickoff(w_in, 1.5, ord(" "))
    with pytest.raises(DSPFatal):
        a_out = np.empty(len_wf)
        inspect.unwrap(fixed_time_pickoff)(w_in, 1.5, ord(" "), a_out)

    # linear tests
    w_in = np.arange(len_wf, dtype=float)
    assert compare_numba_vs_python(fixed_time_pickoff, w_in, 3, ord("i")) == 3

    chars = ["n", "f", "c", "l", "h", "s"]
    sols = [4, 3, 4, 3.5, 3.5, 3.5]

    for char, sol in zip(chars, sols):
        assert compare_numba_vs_python(fixed_time_pickoff, w_in, 3.5, ord(char)) == sol

    # sine wave tests
    w_in = np.sin(np.arange(len_wf))

    chars = ["n", "f", "c", "l", "h", "s"]
    sols = [
        0.1411200080598672,
        0.1411200080598672,
        -0.7568024953079282,
        -0.08336061778208165,
        -0.09054574599004982,
        -0.10707938709427486,
    ]

    for char, sol in zip(chars, sols):
        assert np.isclose(
            compare_numba_vs_python(fixed_time_pickoff, w_in, 3.25, ord(char)), sol
        )

    # last few corner cases of 'h'
    w_in = np.sin(np.arange(len_wf))
    ftps = [0.2, len_wf - 1.8]
    sols = [
        0.1806725096462211,
        -0.6150034250096629,
    ]

    for ftp, sol in zip(ftps, sols):
        assert np.isclose(
            compare_numba_vs_python(fixed_time_pickoff, w_in, ftp, ord("h")), sol
        )



"""
This module provides a routine for running the energy calibration on Th data
"""

from __future__ import annotations

import json
import logging
import math
import os
import pathlib
from datetime import datetime

import matplotlib as mpl
from scipy.stats import binned_statistic

mpl.use("agg")
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LogNorm
from scipy.optimize import curve_fit

import pygama.lgdo.lh5_store as lh5
import pygama.math.histogram as pgh
import pygama.math.peak_fitting as pgf
import pygama.pargen.cuts as cts
import pygama.pargen.energy_cal as cal

log = logging.getLogger(__name__)


def fwhm_slope(x: np.array, m0: float, m1: float, m2: float = None) -> np.array:
    """
    Fit the energy resolution curve
    """
    if m2 is None:
        return np.sqrt(m0 + m1 * x)
    else:
        return np.sqrt(m0 + m1 * x + m2 * x**2)


def load_data(
    files: list[str],
    lh5_path: str,
    energy_params: list[str],
    hit_dict: dict = {},
    cut_parameters: list[str] = ["bl_mean", "bl_std", "pz_std"],
) -> pd.DataFrame:
    df = lh5.load_dfs(files, ["timestamp", "trapTmax"], lh5_path)
    pulser_props = cts.find_pulser_properties(df, energy="trapTmax")
    if len(pulser_props) > 0:
        final_mask = None
        for entry in pulser_props:
            e_cut = (df.trapTmax.values < entry[0] + entry[1]) & (
                df.trapTmax.values > entry[0] - entry[1]
            )
            if final_mask is None:
                final_mask = e_cut
            else:
                final_mask = final_mask | e_cut
        ids = ~(final_mask)
        log.debug(f"pulser found: {pulser_props}")

    else:
        ids = np.ones(len(df), dtype=bool)
        log.debug(f"no pulser found")

    sto = lh5.LH5Store()
    table = sto.read_object(lh5_path, files)[0]

    if len(hit_dict.keys()) == 0:
        out_df = df.copy()
        for param in energy_params:
            try:
                out_df[param] = table[param].nda

            except RuntimeError:
                param = param.split("_")[0]
                out_df[param] = table[param].nda

    else:
        out_df = table.eval(hit_dict).get_dataframe()
        out_df = pd.concat([df, out_df], axis=1)
        out_df["is_not_pulser"] = ids

        cut_parameters = cts.get_keys(table, cut_parameters)

        for param in energy_params:
            if param not in out_df:
                out_df[param] = table[param].nda
        if cut_parameters is not None:
            for param in cut_parameters:
                if param not in df:
                    out_df[param] = table[param].nda
    log.debug("Data Loaded")
    return out_df


def apply_cuts(
    data: pd.DataFrame,
    hit_dict,
    cut_parameters=None,
    final_cut_field: str = "is_valid_cal",
):
    if cut_parameters is not None:
        cut_dict = cts.generate_cuts(data.query("is_not_pulser"), cut_parameters)
        hit_dict.update(
            cts.cut_dict_to_hit_dict(cut_dict, final_cut_field=final_cut_field)
        )
        mask = cts.get_cut_indexes(data, cut_dict)

        data["is_valid_cal"] = mask

    else:
        data["is_valid_cal"] = np.ones(len(data), dtype=bool)
    data["is_usable"] = data["is_valid_cal"] & data["is_not_pulser"]

    events_pqc = len(data.query("is_usable"))
    log.debug(f"{events_pqc} events valid for calibration")

    return data, hit_dict


def gen_pars_dict(pars, deg, energy_param):
    if deg == 1:
        out_dict = {
            "expression": f"a*{energy_param}+b",
            "parameters": {"a": pars[0], "b": pars[1]},
        }
    elif deg == 0:
        out_dict = {
            "expression": f"a*{energy_param}",
            "parameters": {"a": pars[0]},
        }
    elif deg == 2:
        out_dict = {
            "expression": f"a*{energy_param}**2 +b*{energy_param}+c",
            "parameters": {"a": pars[0], "b": pars[1], "c": pars[2]},
        }
    else:
        out_dict = {}
        log.error(f"hit_dict not implemented for deg = {deg}")

    return out_dict


class calibrate_parameter:
    glines = [
        # 238.632,
        583.191,
        727.330,
        860.564,
        1592.53,
        1620.50,
        2103.53,
        2614.50,
    ]  # gamma lines used for calibration
    range_keV = [
        # (8, 8),
        (20, 20),
        (30, 30),
        (30, 30),
        (40, 25),
        (25, 40),
        (40, 40),
        (60, 60),
    ]  # side bands width
    funcs = [
        # pgf.extended_gauss_step_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
    ]
    gof_funcs = [
        # pgf.gauss_step_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
    ]

    def __init__(
        self,
        data,
        energy_param,
        plot_options: dict = None,
        guess_keV: float | None = None,
        threshold: int = 0,
        p_val: float = 0,
        n_events: int = 15000,
        deg: int = 1,
    ):
        self.data = data
        self.energy_param = energy_param
        self.guess_keV = guess_keV
        self.threshold = threshold
        self.p_val = p_val
        self.n_events = n_events
        self.deg = deg
        self.plot_options = plot_options

        self.output_dict = {}
        self.hit_dict = {}
        self.plot_dict = {}

    def fit_energy_res(self):
        fitted_peaks = self.results["fitted_keV"]
        fwhms = self.results["pk_fwhms"][:, 0]
        dfwhms = self.results["pk_fwhms"][:, 1]

        #####
        # Remove the Tl SEP and DEP from calibration if found
        fwhm_peaks = np.array([], dtype=np.float32)
        all_peaks = np.array([], dtype=np.float32)
        indexes = []
        for i, peak in enumerate(fitted_peaks):
            all_peaks = np.append(all_peaks, peak)
            if peak == 2103.53:
                log.info(f"Tl SEP found at index {i}")
                indexes.append(i)
                continue
            elif peak == 1592.53:
                log.info(f"Tl DEP found at index {i}")
                indexes.append(i)
                continue
            elif np.isnan(dfwhms[i]):
                log.info(f"{peak} failed")
                indexes.append(i)
                continue
            else:
                fwhm_peaks = np.append(fwhm_peaks, peak)
        fit_fwhms = np.delete(fwhms, [indexes])
        fit_dfwhms = np.delete(dfwhms, [indexes])
        #####
        param_guess = [2, 0.001]
        param_bounds = (0, np.inf)
        for i, peak in enumerate(fwhm_peaks):
            log.info(
                f"FWHM of {peak} keV peak is: {fit_fwhms[i]:1.2f} +- {fit_dfwhms[i]:1.2f} keV"
            )
        try:
            self.fit_pars, self.fit_covs = curve_fit(
                fwhm_slope,
                fwhm_peaks,
                fit_fwhms,
                sigma=fit_dfwhms,
                p0=param_guess,
                bounds=param_bounds,
                absolute_sigma=True,
            )
            rng = np.random.default_rng(1)
            pars_b = rng.multivariate_normal(self.fit_pars, self.fit_covs, size=1000)
            fits = np.array([fwhm_slope(fwhm_peaks, *par_b) for par_b in pars_b])
            qbb_vals = np.array([fwhm_slope(2039.0, *par_b) for par_b in pars_b])
            self.qbb_err = np.nanstd(qbb_vals)
            predicted_fwhms = fwhm_slope(fwhm_peaks, *self.fit_pars)
            self.fit_qbb = fwhm_slope(2039.0, *self.fit_pars)

            if 2614.50 not in fwhm_peaks:
                self.fit_qbb = np.nan
                self.qbb_err = np.nan
            log.info(f"FWHM curve fit: {self.fit_pars}")
            log.info(f"FWHM fit values:")
            for peak in fwhm_peaks:
                log.info(
                    f"Predicted FWHM of {peak} keV peak is: {fwhm_slope(peak, *self.fit_pars):.2f} keV"
                )
            log.info(
                f"FWHM energy resolution at Qbb: {self.fit_qbb:1.2f} +- {self.qbb_err:1.2f} keV"
            )
        except RuntimeError:
            log.error(f"FWHM fit failed for {energy_param}")
            self.fit_pars = np.array([np.nan, np.nan])
            self.fit_covs = np.array([[np.nan, np.nan], [np.nan, np.nan]])
            self.fit_qbb = np.nan
            self.qbb_err = np.nan
            log.error("FWHM fit failed to converge")

    def gen_pars_dict(self):
        if self.deg == 1:
            out_dict = {
                "expression": f"a*{self.energy_param}+b",
                "parameters": {"a": self.pars[0], "b": self.pars[1]},
            }
        elif self.deg == 0:
            out_dict = {
                "expression": f"a*{self.energy_param}",
                "parameters": {"a": self.pars[0]},
            }
        elif self.deg == 2:
            out_dict = {
                "expression": f"a*{self.energy_param}**2 +b*{self.energy_param}+c",
                "parameters": {"a": self.pars[0], "b": self.pars[1], "c": self.pars[2]},
            }
        else:
            out_dict = {}
            log.warning(f"hit_dict not implemented for deg = {self.deg}")

        return out_dict

    def calibrate_parameter(self):
        kev_ranges = self.range_keV.copy()
        if self.guess_keV is None:
            self.guess_keV = 2620 / np.nanpercentile(
                self.data.query(f"is_usable & {self.energy_param}>{self.threshold}")[
                    self.energy_param
                ],
                99,
            )

        log.debug(f"Find peaks and compute calibration curve for {self.energy_param}")
        log.debug(f"Guess is {self.guess_keV:.3f}")

        try:
            self.pars, self.cov, self.results = cal.hpge_E_calibration(
                self.data.query("is_usable")[self.energy_param],
                self.glines,
                self.guess_keV,
                deg=self.deg,
                range_keV=kev_ranges,
                funcs=self.funcs,
                gof_funcs=self.gof_funcs,
                n_events=self.n_events,
                allowed_p_val=self.p_val,
                simplex=True,
                verbose=False,
            )
            pk_pars = self.results["pk_pars"]
            found_peaks = self.results["got_peaks_locs"]
            fitted_peaks = self.results["fitted_keV"]
        except:
            found_peaks = np.array([])
            fitted_peaks = np.array([])

        for i, peak in enumerate(self.glines):
            if peak not in fitted_peaks:
                kev_ranges[i] = (kev_ranges[i][0] - 5, kev_ranges[i][1] - 5)
        for i, peak in enumerate(self.glines):
            if peak not in fitted_peaks:
                kev_ranges[i] = (kev_ranges[i][0] - 5, kev_ranges[i][1] - 5)
        for i, peak in enumerate(fitted_peaks):
            try:
                if (
                    self.results["pk_fwhms"][:, 1][i]
                    / self.results["pk_fwhms"][:, 0][i]
                    > 0.05
                ):
                    index = np.where(self.glines == peak)[0][0]
                    kev_ranges[i] = (kev_ranges[index][0] - 5, kev_ranges[index][1] - 5)
            except:
                pass

        try:
            self.pars, self.cov, self.results = cal.hpge_E_calibration(
                self.data.query("is_usable")[self.energy_param],
                self.glines,
                self.guess_keV,
                deg=self.deg,
                range_keV=kev_ranges,
                funcs=self.funcs,
                gof_funcs=self.gof_funcs,
                n_events=self.n_events,
                allowed_p_val=self.p_val,
                simplex=True,
                verbose=False,
            )
        except:
            self.pars = None
        if self.pars is None:
            log.error(
                f"Calibration failed for {self.energy_param}, trying with 0 p_val"
            )
            try:
                self.pars, self.cov, self.results = cal.hpge_E_calibration(
                    self.data.query("is_usable")[self.energy_param],
                    self.glines,
                    self.guess_keV,
                    deg=self.deg,
                    range_keV=kev_ranges,
                    funcs=self.funcs,
                    gof_funcs=self.gof_funcs,
                    n_events=self.n_events,
                    allowed_p_val=0,
                    simplex=True,
                    verbose=False,
                )
                if self.pars is None:
                    raise ValueError

                self.fit_energy_res()
                self.data[f"{self.energy_param}_cal"] = pgf.poly(
                    self.data[self.energy_param], self.pars
                )
                self.hit_dict[f"{self.energy_param}_cal"] = self.gen_pars_dict()
                self.output_dict[f"{self.energy_param}_cal"] = {
                    "Qbb_fwhm": np.nan,
                    "Qbb_fwhm_err": np.nan,
                    "2.6_fwhm": np.nan,
                    "2.6_fwhm_err": np.nan,
                    "eres_pars": self.fit_pars.tolist(),
                    "fitted_peaks": np.nan,
                    "fwhms": np.nan,
                    "peak_fit_pars": np.nan,
                    "total_fep": len(
                        self.data.query(
                            f"{self.energy_param}_cal>2604&{self.energy_param}_cal<2624"
                        )
                    ),
                    "total_dep": len(
                        self.data.query(
                            f"{self.energy_param}_cal>1587&{self.energy_param}_cal<1597"
                        )
                    ),
                    "pass_fep": len(
                        self.data.query(
                            f"{self.energy_param}_cal>2604&{self.energy_param}_cal<2624&is_usable"
                        )
                    ),
                    "pass_dep": len(
                        self.data.query(
                            f"{self.energy_param}_cal>1587&{self.energy_param}_cal<1597&is_usable"
                        )
                    ),
                }
            except:
                log.error(
                    f"Calibration failed completely for {self.energy_param} even with 0 p_val"
                )
                self.pars = np.full(self.deg + 1, np.nan)

                self.hit_dict[f"{self.energy_param}_cal"] = self.gen_pars_dict()

                self.output_dict[f"{self.energy_param}_cal"] = {
                    "Qbb_fwhm": np.nan,
                    "Qbb_fwhm_err": np.nan,
                    "2.6_fwhm": np.nan,
                    "2.6_fwhm_err": np.nan,
                    "eres_pars": [np.nan, np.nan],
                    "fitted_peaks": np.nan,
                    "fwhms": np.nan,
                    "peak_fit_pars": np.nan,
                    "total_fep": np.nan,
                    "total_dep": np.nan,
                    "pass_fep": np.nan,
                    "pass_dep": np.nan,
                }

        else:
            log.debug("done")
            log.info(f"Calibration pars are {self.pars}")

            self.data[f"{self.energy_param}_cal"] = pgf.poly(
                self.data[self.energy_param], self.pars
            )

            pk_rs_dict = {
                peak: self.results["pk_pars"][i].tolist()
                for i, peak in enumerate(self.results["fitted_keV"])
            }

            self.fit_energy_res()
            self.hit_dict[f"{self.energy_param}_cal"] = self.gen_pars_dict()

            if self.results["fitted_keV"][-1] == 2614.50:
                fep_fwhm = round(self.results["pk_fwhms"][-1, 0], 2)
                fep_dwhm = round(self.results["pk_fwhms"][-1, 1], 2)
            else:
                fep_fwhm = np.nan
                fep_dwhm = np.nan

            self.output_dict[f"{self.energy_param}_cal"] = {
                "Qbb_fwhm": round(self.fit_qbb, 2),
                "Qbb_fwhm_err": round(self.qbb_err, 2),
                "2.6_fwhm": fep_fwhm,
                "2.6_fwhm_err": fep_dwhm,
                "eres_pars": self.fit_pars.tolist(),
                "fitted_peaks": self.results["fitted_keV"].tolist(),
                "fwhms": self.results["pk_fwhms"].tolist(),
                "peak_fit_pars": pk_rs_dict,
                "total_fep": len(
                    self.data.query(
                        f"{self.energy_param}_cal>2604&{self.energy_param}_cal<2624"
                    )
                ),
                "total_dep": len(
                    self.data.query(
                        f"{self.energy_param}_cal>1587&{self.energy_param}_cal<1597"
                    )
                ),
                "pass_fep": len(
                    self.data.query(
                        f"{self.energy_param}_cal>2604&{self.energy_param}_cal<2624&is_usable"
                    )
                ),
                "pass_dep": len(
                    self.data.query(
                        f"{self.energy_param}_cal>1587&{self.energy_param}_cal<1597&is_usable"
                    )
                ),
            }
        log.info(
            f"Results {self.energy_param}: {json.dumps(self.output_dict[f'{self.energy_param}_cal'], indent=2)}"
        )

    def fill_plot_dict(self):
        for key, item in self.plot_options.items():
            if item["options"] is not None:
                self.plot_dict[key] = item["function"](self, **item["options"])
            else:
                self.plot_dict[key] = item["function"](self)


def get_peak_labels(
    labels: list[str], pars: list[float]
) -> tuple(list[float], list[float]):
    out = []
    out_labels = []
    for i, label in enumerate(labels):
        if i % 2 == 1:
            continue
        else:
            out.append(f"{pgf.poly(label, pars):.1f}")
            out_labels.append(label)
    return out_labels, out


def get_peak_label(peak: float) -> str:
    if peak == 583.191:
        return "Tl 583"
    elif peak == 727.33:
        return "Bi 727"
    elif peak == 860.564:
        return "Tl 860"
    elif peak == 1592.53:
        return "Tl DEP"
    elif peak == 1620.5:
        return "Bi FEP"
    elif peak == 2103.53:
        return "Tl SEP"
    elif peak == 2614.5:
        return "Tl FEP"


def plot_fits(ecal_class, figsize=[12, 8], fontsize=12, ncols=3, n_rows=3):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fitted_peaks = ecal_class.results["fitted_keV"]
    pk_pars = ecal_class.results["pk_pars"]
    pk_ranges = ecal_class.results["pk_ranges"]
    p_vals = ecal_class.results["pk_pvals"]

    fitted_gof_funcs = []
    for i, peak in enumerate(ecal_class.glines):
        if peak in fitted_peaks:
            fitted_gof_funcs.append(ecal_class.gof_funcs[i])

    mus = [
        pgf.get_mu_func(func_i, pars_i)
        for func_i, pars_i in zip(fitted_gof_funcs, pk_pars)
    ]
    fwhms = ecal_class.results["pk_fwhms"][:, 0]
    dfwhms = ecal_class.results["pk_fwhms"][:, 1]

    fig = plt.figure()
    range_adu = 5 / ecal_class.pars[0]  # 10keV window around peak in adu
    for i, peak in enumerate(mus):
        # plt.subplot(math.ceil((len(mus)) / 2), 2, i + 1)
        plt.subplot(n_rows, ncols, i + 1)
        binning = np.arange(pk_ranges[i][0], pk_ranges[i][1], 1)
        bin_cs = (binning[1:] + binning[:-1]) / 2
        energies = ecal_class.data.query(
            f"{ecal_class.energy_param}>{pk_ranges[i][0]}&{ecal_class.energy_param}<{pk_ranges[i][1]}&is_usable"
        )[ecal_class.energy_param]
        energies = energies.iloc[: ecal_class.n_events]

        counts, bs, bars = plt.hist(energies, bins=binning, histtype="step")
        fit_vals = fitted_gof_funcs[i](bin_cs, *pk_pars[i]) * np.diff(bs)
        plt.plot(bin_cs, fit_vals)
        plt.step(
            bin_cs,
            [
                (fval - count) / count if count != 0 else (fval - count)
                for count, fval in zip(counts, fit_vals)
            ],
        )
        plt.plot(
            [bin_cs[10]],
            [0],
            label=get_peak_label(fitted_peaks[i]),
            linestyle="None",
        )
        plt.plot(
            [bin_cs[10]],
            [0],
            label=f"{fitted_peaks[i]:.1f} keV",
            linestyle="None",
        )
        plt.plot(
            [bin_cs[10]],
            [0],
            label=f"{fwhms[i]:.2f} +- {dfwhms[i]:.2f} keV",
            linestyle="None",
        )
        plt.plot(
            [bin_cs[10]],
            [0],
            label=f"p-value : {p_vals[i]:.2f}",
            linestyle="None",
        )

        plt.xlabel("Energy (keV)")
        plt.ylabel("Counts")
        plt.legend(loc="upper left", frameon=False)
        plt.xlim([peak - range_adu, peak + range_adu])
        locs, labels = plt.xticks()
        new_locs, new_labels = get_peak_labels(locs, ecal_class.pars)
        plt.xticks(ticks=new_locs, labels=new_labels)

    plt.tight_layout()
    plt.close()
    return fig


def plot_2614_timemap(
    ecal_class, figsize=[12, 8], fontsize=12, erange=[2580, 2630], dx=1, time_dx=180
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    selection = ecal_class.data.query(
        f"{ecal_class.energy_param}_cal>2560&{ecal_class.energy_param}_cal<2660&is_usable"
    )

    if len(selection) == 0:
        pass
    else:
        time_bins = np.arange(
            (np.amin(ecal_class.data["timestamp"]) // time_dx) * time_dx,
            ((np.amax(ecal_class.data["timestamp"]) // time_dx) + 2) * time_dx,
            time_dx,
        )

        fig = plt.figure()
        plt.hist2d(
            selection["timestamp"],
            selection[f"{ecal_class.energy_param}_cal"],
            bins=[time_bins, np.arange(erange[0], erange[1] + dx, dx)],
            norm=LogNorm(),
        )

    ticks, labels = plt.xticks()
    plt.xlabel(
        f"Time starting : {datetime.utcfromtimestamp(ticks[0]).strftime('%d/%m/%y %H:%M')}"
    )
    plt.ylabel("Energy(keV)")
    plt.ylim([erange[0], erange[1]])

    plt.xticks(
        ticks,
        [datetime.utcfromtimestamp(tick).strftime("%H:%M") for tick in ticks],
    )
    plt.close()
    return fig


def plot_pulser_timemap(
    ecal_class, figsize=[12, 8], fontsize=12, dx=0.2, time_dx=180, n_spread=3
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    time_bins = np.arange(
        (np.amin(ecal_class.data["timestamp"]) // time_dx) * time_dx,
        ((np.amax(ecal_class.data["timestamp"]) // time_dx) + 2) * time_dx,
        time_dx,
    )

    selection = ecal_class.data.query(f"~is_not_pulser")
    fig = plt.figure()
    if len(selection) == 0:
        pass

    else:
        mean = np.nanpercentile(selection[f"{ecal_class.energy_param}_cal"], 50)
        spread = mean - np.nanpercentile(
            selection[f"{ecal_class.energy_param}_cal"], 10
        )

        plt.hist2d(
            selection["timestamp"],
            selection[f"{ecal_class.energy_param}_cal"],
            bins=[
                time_bins,
                np.arange(mean - n_spread * spread, mean + n_spread * spread + dx, dx),
            ],
            norm=LogNorm(),
        )
        plt.ylim([mean - n_spread * spread, mean + n_spread * spread])
    ticks, labels = plt.xticks()
    plt.xlabel(
        f"Time starting : {datetime.utcfromtimestamp(ticks[0]).strftime('%d/%m/%y %H:%M')}"
    )
    plt.ylabel("Energy(keV)")

    plt.xticks(
        ticks,
        [datetime.utcfromtimestamp(tick).strftime("%H:%M") for tick in ticks],
    )
    plt.close()
    return fig


def bin_pulser_stability(ecal_class, time_slice=180):
    selection = ecal_class.data.query(f"~is_not_pulser")

    utime_array = ecal_class.data["timestamp"]
    select_energies = selection[f"{ecal_class.energy_param}_cal"].to_numpy()

    time_bins = np.arange(
        (np.amin(utime_array) // time_slice) * time_slice,
        ((np.amax(utime_array) // time_slice) + 2) * time_slice,
        time_slice,
    )
    # bin time values
    times_average = (time_bins[:-1] + time_bins[1:]) / 2

    if len(selection) == 0:
        return {
            "time": times_average,
            "energy": np.full_like(times_average, np.nan),
            "spread": np.full_like(times_average, np.nan),
        }

    nanmedian = (
        lambda x: np.nanpercentile(x, 50) if len(x[~np.isnan(x)]) >= 10 else np.nan
    )
    error = (
        lambda x: np.nanvar(x) / np.sqrt(len(x))
        if len(x[~np.isnan(x)]) >= 10
        else np.nan
    )

    par_average, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=nanmedian, bins=time_bins
    )
    par_error, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=error, bins=time_bins
    )

    return {"time": times_average, "energy": par_average, "spread": par_error}


def bin_stability(ecal_class, time_slice=180, energy_range=[2585, 2660]):
    selection = ecal_class.data.query(
        f"{ecal_class.energy_param}_cal>{energy_range[0]}&{ecal_class.energy_param}_cal<{energy_range[1]}&is_usable"
    )

    utime_array = ecal_class.data["timestamp"]
    select_energies = selection[f"{ecal_class.energy_param}_cal"].to_numpy()

    time_bins = np.arange(
        (np.amin(utime_array) // time_slice) * time_slice,
        ((np.amax(utime_array) // time_slice) + 2) * time_slice,
        time_slice,
    )
    # bin time values
    times_average = (time_bins[:-1] + time_bins[1:]) / 2

    if len(selection) == 0:
        return {
            "time": times_average,
            "energy": np.full_like(times_average, np.nan),
            "spread": np.full_like(times_average, np.nan),
        }

    nanmedian = (
        lambda x: np.nanpercentile(x, 50) if len(x[~np.isnan(x)]) >= 10 else np.nan
    )
    error = (
        lambda x: np.nanvar(x) / np.sqrt(len(x))
        if len(x[~np.isnan(x)]) >= 10
        else np.nan
    )

    par_average, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=nanmedian, bins=time_bins
    )
    par_error, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=error, bins=time_bins
    )

    return {"time": times_average, "energy": par_average, "spread": par_error}


def plot_cal_fit(ecal_class, figsize=[12, 8], fontsize=12, erange=[200, 2700]):
    pk_pars = ecal_class.results["pk_pars"]
    fitted_peaks = ecal_class.results["fitted_keV"]

    fitted_gof_funcs = []
    for i, peak in enumerate(ecal_class.glines):
        if peak in fitted_peaks:
            fitted_gof_funcs.append(ecal_class.gof_funcs[i])

    mus = [
        pgf.get_mu_func(func_i, pars_i)
        for func_i, pars_i in zip(fitted_gof_funcs, pk_pars)
    ]

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig, (ax1, ax2) = plt.subplots(
        2, 1, sharex=True, gridspec_kw={"height_ratios": [3, 1]}
    )

    cal_bins = np.arange(0, np.nanmax(mus) * 1.1, 10)

    ax1.scatter(fitted_peaks, mus, marker="x", c="b")

    ax1.plot(pgf.poly(cal_bins, ecal_class.pars), cal_bins, lw=1, c="g")

    ax1.grid()
    ax1.set_xlim([erange[0], erange[1]])
    ax1.set_ylabel("Energy (ADC)")
    ax2.plot(
        fitted_peaks,
        pgf.poly(np.array(mus), ecal_class.pars) - fitted_peaks,
        lw=0,
        marker="x",
        c="b",
    )
    ax2.grid()
    ax2.set_xlabel("Energy (keV)")
    ax2.set_ylabel("Residuals (keV)")
    plt.close()
    return fig


def plot_eres_fit(ecal_class, figsize=[12, 8], fontsize=12, erange=[200, 2700]):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fwhms = ecal_class.results["pk_fwhms"][:, 0]
    dfwhms = ecal_class.results["pk_fwhms"][:, 1]
    fitted_peaks = ecal_class.results["fitted_keV"]

    #####
    # Remove the Tl SEP and DEP from calibration if found
    fwhm_peaks = np.array([], dtype=np.float32)
    indexes = []
    for i, peak in enumerate(fitted_peaks):
        if peak == 2103.53:
            log.info(f"Tl SEP found at index {i}")
            indexes.append(i)
            continue
        elif peak == 1592.53:
            log.info(f"Tl DEP found at index {i}")
            indexes.append(i)
            continue
        elif np.isnan(dfwhms[i]):
            log.info(f"{peak} failed")
            indexes.append(i)
            continue
        else:
            fwhm_peaks = np.append(fwhm_peaks, peak)
    fit_fwhms = np.delete(fwhms, [indexes])
    fit_dfwhms = np.delete(dfwhms, [indexes])

    fig, (ax1, ax2) = plt.subplots(
        2, 1, sharex=True, gridspec_kw={"height_ratios": [3, 1]}
    )
    ax1.errorbar(fwhm_peaks, fit_fwhms, yerr=fit_dfwhms, marker="x", lw=0, c="b")

    fwhm_slope_bins = np.arange(erange[0], erange[1], 10)

    qbb_line_vx = [2039.0, 2039.0]
    qbb_line_vy = [
        0.9 * np.nanmin(fwhm_slope(fwhm_slope_bins, *ecal_class.fit_pars)),
        ecal_class.fit_qbb,
    ]
    qbb_line_hx = [erange[0], 2039.0]
    qbb_line_hy = [ecal_class.fit_qbb, ecal_class.fit_qbb]

    ax1.plot(
        fwhm_slope_bins, fwhm_slope(fwhm_slope_bins, *ecal_class.fit_pars), lw=1, c="g"
    )
    ax1.plot(qbb_line_hx, qbb_line_hy, lw=1, c="r")
    ax1.plot(qbb_line_vx, qbb_line_vy, lw=1, c="r")
    ax1.plot(
        np.nan,
        np.nan,
        "-",
        color="none",
        label=f"Qbb fwhm: {ecal_class.fit_qbb:1.2f} +- {ecal_class.qbb_err:1.2f} keV",
    )
    ax1.legend(loc="upper left", frameon=False)
    if np.isnan(ecal_class.fit_pars).all():
        [
            0.9 * np.nanmin(fit_fwhms),
            1.1 * np.nanmax(fit_fwhms),
        ]
    else:
        ax1.set_ylim(
            [
                0.9 * np.nanmin(fwhm_slope(fwhm_slope_bins, *ecal_class.fit_pars)),
                1.1 * np.nanmax(fwhm_slope(fwhm_slope_bins, *ecal_class.fit_pars)),
            ]
        )
    ax1.set_xlim([200, 2700])
    ax1.grid()
    ax1.set_ylabel("FWHM energy resolution (keV)")
    ax2.plot(
        fwhm_peaks,
        (fit_fwhms - fwhm_slope(fwhm_peaks, *ecal_class.fit_pars)) / fit_dfwhms,
        lw=0,
        marker="x",
        c="b",
    )
    ax2.set_xlabel("Energy (keV)")
    ax2.set_ylabel("Normalised Residuals")
    ax2.grid()
    plt.tight_layout()
    plt.close()
    return fig


def bin_spectrum(ecal_class, erange=[0, 3000], dx=2):
    bins = np.arange(erange[0], erange[1] + dx, dx)
    return {
        "bins": pgh.get_bin_centers(bins),
        "counts": np.histogram(
            ecal_class.data.query("is_usable")[f"{ecal_class.energy_param}_cal"], bins
        )[0],
        "cut_counts": np.histogram(
            ecal_class.data.query("~is_valid_cal&is_not_pulser")[
                f"{ecal_class.energy_param}_cal"
            ],
            bins,
        )[0],
        "pulser_counts": np.histogram(
            ecal_class.data.query("~is_not_pulser")[f"{ecal_class.energy_param}_cal"],
            bins,
        )[0],
    }


def bin_survival_fraction(ecal_class, erange=[0, 3000], dx=6):
    counts_pass, bins_pass, _ = pgh.get_hist(
        ecal_class.data.query("is_usable")[f"{ecal_class.energy_param}_cal"],
        bins=np.arange(erange[0], erange[1] + dx, dx),
    )
    counts_fail, bins_fail, _ = pgh.get_hist(
        ecal_class.data.query("~is_valid_cal&is_not_pulser")[
            f"{ecal_class.energy_param}_cal"
        ],
        bins=np.arange(erange[0], erange[1] + dx, dx),
    )
    sf = 100 * (counts_pass + 10 ** (-6)) / (counts_pass + counts_fail + 10 ** (-6))
    return {"bins": pgh.get_bin_centers(bins_pass), "sf": sf}


def energy_cal_th(
    files: list[str],
    energy_params: list[str],
    hit_dict: dict = {},
    cut_parameters: dict[str, int] = {"bl_mean": 4, "bl_std": 4, "pz_std": 4},
    lh5_path: str = "dsp",
    plot_options: dict = None,
    guess_keV: float | None = None,
    threshold: int = 0,
    p_val: float = 0,
    n_events: int = 15000,
    final_cut_field: str = "is_valid_cal",
    deg: int = 1,
) -> tuple(dict, dict):
    data = load_data(
        files,
        lh5_path,
        energy_params,
        hit_dict,
        cut_parameters=list(cut_parameters) if cut_parameters is not None else None,
    )

    data, hit_dict = apply_cuts(data, hit_dict, cut_parameters, final_cut_field)

    output_dict = {}
    plot_dict = {}
    for energy_param in energy_params:
        ecal = calibrate_parameter(
            data, energy_param, plot_options, guess_keV, threshold, p_val, n_events, deg
        )
        ecal.calibrate_parameter()
        output_dict.update(ecal.output_dict)
        hit_dict.update(ecal.hit_dict)
        if ~np.isnan(ecal.pars).all():
            ecal.fill_plot_dict()
        plot_dict[energy_param] = ecal.plot_dict

    log.info(f"Finished all calibrations")
    return hit_dict, output_dict, plot_dict



import os
from collections.abc import MutableMapping
from typing import Any, Iterator


def getenv_bool(name: str, default: bool = False) -> bool:
    """Get environment value as a boolean, returning True for 1, t and true
    (caps-insensitive), and False for any other value and default if undefined.
    """
    val = os.getenv(name)
    if not val:
        return default
    elif val.lower() in ("1", "t", "true"):
        return True
    else:
        return False


class NumbaDefaults(MutableMapping):
    """Bare-bones class to store some Numba default options. Defaults values
    are set from environment variables

    Examples
    --------
    Set all default option values for a processor at once by expanding the
    provided dictionary:

    >>> from numba import guvectorize
    >>> from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs
    >>> @guvectorize([], "", **nb_kwargs, nopython=True) # def proc(...): ...

    Customize one argument but still set defaults for the others:

    >>> from pygama.dsp.utils import numba_defaults as nb_defaults
    >>> @guvectorize([], "", **nb_defaults(cache=False) # def proc(...): ...

    Override global options at runtime:

    >>> from pygama.dsp.utils import numba_defaults
    >>> from pygama.dsp import build_dsp
    >>> # must set options before explicitly importing pygama.dsp.processors!
    >>> numba_defaults.cache = False
    >>> numba_defaults.boundscheck = True
    >>> build_dsp(...) # if not explicit, processors imports happen here
    """

    def __init__(self) -> None:
        self.cache: bool = getenv_bool("PYGAMA_CACHE")
        self.boundscheck: bool = getenv_bool("PYGAMA_BOUNDSCHECK")

    def __getitem__(self, item: str) -> Any:
        return self.__dict__[item]

    def __setitem__(self, item: str, val: Any) -> None:
        self.__dict__[item] = val

    def __delitem__(self, item: str) -> None:
        del self.__dict__[item]

    def __iter__(self) -> Iterator:
        return self.__dict__.__iter__()

    def __len__(self) -> int:
        return len(self.__dict__)

    def __call__(self, **kwargs) -> dict:
        mapping = self.__dict__.copy()
        mapping.update(**kwargs)
        return mapping

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return str(self.__dict__)


numba_defaults = NumbaDefaults()
numba_defaults_kwargs = numba_defaults



import copy
import inspect
import os
import re
from pathlib import Path

import numpy as np
import pytest
from legend_testdata import LegendTestData

from pygama.dsp import build_dsp
from pygama.raw import build_raw

config_dir = Path(__file__).parent / "dsp" / "configs"


@pytest.fixture(scope="session")
def lgnd_test_data():
    ldata = LegendTestData()
    ldata.checkout("39f9927")
    return ldata


@pytest.fixture(scope="session")
def dsp_test_file(lgnd_test_data):
    out_name = "/tmp/LDQTA_r117_20200110T105115Z_cal_geds_dsp.lh5"
    build_dsp(
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        out_name,
        dsp_config=f"{config_dir}/icpc-dsp-config.json",
        database={"pz": {"tau": 27460.5}},
        write_mode="r",
    )
    assert os.path.exists(out_name)

    return out_name


@pytest.fixture(scope="session")
def multich_raw_file(lgnd_test_data):
    out_file = "/tmp/L200-comm-20211130-phy-spms.lh5"
    out_spec = {
        "FCEventDecoder": {
            "ch{key}": {
                "key_list": [[0, 6]],
                "out_stream": out_file + ":{name}",
                "out_name": "raw",
            }
        }
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=out_spec,
        overwrite=True,
    )
    assert os.path.exists(out_file)

    return out_file


@pytest.fixture(scope="session")
def dsp_test_file_spm(multich_raw_file):
    chan_config = {
        "ch0/raw": f"{config_dir}/sipm-dsp-config.json",
        "ch1/raw": f"{config_dir}/sipm-dsp-config.json",
        "ch2/raw": f"{config_dir}/sipm-dsp-config.json",
    }

    out_file = "/tmp/L200-comm-20211130-phy-spms_dsp.lh5"
    build_dsp(
        multich_raw_file,
        out_file,
        {},
        n_max=5,
        lh5_tables=chan_config.keys(),
        chan_config=chan_config,
        write_mode="r",
    )

    assert os.path.exists(out_file)

    return out_file


@pytest.fixture(scope="session")
def compare_numba_vs_python():
    def numba_vs_python(func, *inputs):
        """
        Function for testing that the numba and python versions of a
        function are equal.

        Parameters
        ----------
        func
            The Numba-wrapped function to be tested
        *inputs
            The various inputs to be passed to the function to be
            tested.

        Returns
        -------
        func_output
            The output of the function to be used in a unit test.

        """

        if "->" in func.signature:
            # parse outputs from function signature
            all_params = list(inspect.signature(func).parameters)
            output_sizes = re.findall(r"(\(n*\))", func.signature.split("->")[-1])
            noutputs = len(output_sizes)
            output_names = all_params[-noutputs:]

            # numba outputs
            outputs_numba = func(*inputs)
            if noutputs == 1:
                outputs_numba = [outputs_numba]

            # unwrapped python outputs
            func_unwrapped = inspect.unwrap(func)
            output_dict = {key: np.empty(len(inputs[0])) for key in output_names}
            func_unwrapped(*inputs, **output_dict)
            for spec, key in zip(output_sizes, output_dict):
                if spec == "()":
                    output_dict[key] = output_dict[key][0]
            outputs_python = [output_dict[key] for key in output_names]
        else:
            # we are testing a factory function output, which updates
            # a single output in-place
            noutputs = 1
            # numba outputs
            func(*inputs)
            outputs_numba = copy.deepcopy(inputs[-noutputs:])

            # unwrapped python outputs
            func_unwrapped = inspect.unwrap(func)
            func_unwrapped(*inputs)
            outputs_python = copy.deepcopy(inputs[-noutputs:])

        # assert that numba and python are the same up to floating point
        # precision, setting nans to be equal
        assert np.allclose(outputs_numba, outputs_python, equal_nan=True)

        # return value for comparison with expected solution
        return outputs_numba[0] if noutputs == 1 else outputs_numba

    return numba_vs_python



from pint import Measurement, Quantity, Unit, UnitRegistry, set_application_registry

unit_registry = UnitRegistry(auto_reduce_dimensions=True)
set_application_registry(unit_registry)

# Define constants useful for LEGEND below
unit_registry.define('m_76 = 75.921402729 * amu') # 10.1103/PhysRevC.81.032501
unit_registry.define('Q_bb = 2039.061 * keV') # 10.1103/PhysRevC.81.032501



"""
Base classes for decoding data into raw LGDO Tables or files
"""
from __future__ import annotations

from typing import Union

import numpy as np

from pygama import lgdo
from pygama.lgdo.lh5_store import LH5Store
from pygama.raw.raw_buffer import RawBuffer

LGDO = Union[lgdo.Scalar, lgdo.Struct, lgdo.Array, lgdo.VectorOfVectors]


class DataDecoder:
    r"""Decodes packets from a data stream.

    Most decoders will repeatedly decode the same set of values from each
    packet.  The values that get decoded need to be described by a dict stored
    in `self.decoded_values` that helps determine how to set up the buffers and
    write them to file as :class:`~.lgdo.LGDO`\ s. :class:`~.lgdo.table.Table`\ s
    are made whose columns correspond to the elements of `decoded_values`, and
    packet data gets pushed to the end of the table one row at a time.

    Any key-value entry in a configuration dictionary attached to an element
    of `decoded_values` is typically interpreted as an attribute to be attached
    to the corresponding LGDO. This feature can be for example exploited to
    specify the data compression algorithm used by
    :meth:`~.lgdo.lh5_store.LH5Store.write_object` to write LGDOs to disk.

    For example ::

      from pygama.lgdo.compression import RadwareSigcompress

      FCEventDecoder.decoded_values = {
        "packet_id": {"dtype": "uint32", "compression": "gzip"},
        # ...
        "waveform": {
          "dtype": "uint16",
          "datatype": "waveform",
          # ...
          "compression": {"values": RadwareSigcompress(codec_shift=-32768)},
        }
      }

    LGDOs corresponding to ``packet_id`` and ``waveform`` will have their
    `compression` attribute set as ``"gzip"`` and
    ``RadwareSigcompress(codec_shift=-32768)``, respectively. Before being
    written to disk, they will compressed with the HDF5 built-in Gzip filter
    and with the :class:`~.lgdo.compression.radware.RadwareSigcompress`
    waveform compressor.

    Examples
    --------
    See `decoded_values` attributes of
    :class:`~.fc.fc_event_decoder.FCEventDecoder` or
    :class:`~.orca.orca_digitizers.ORSIS3316WaveformDecoder`.

    Some decoders (like for file headers) do not need to push to a table, so they
    do not need `decoded_values`. Such classes should still derive from
    :class:`DataDecoder` and define how data gets formatted into LGDO's.

    Subclasses should define a method for decoding data to a buffer like
    ``decode_packet(packet, raw_buffer_list, packet_id)``.  This function
    should return the number of bytes read.

    Garbage collection writes binary data as an array of :obj:`~numpy.uint32`\ s
    to a variable-length array in the output file. If a problematic packet is
    found, call :meth:`.put_in_garbage`. User should set up an enum or bitbank
    of garbage codes to be stored along with the garbage packets.
    """

    def __init__(
        self, garbage_length: int = 256, packet_size_guess: int = 1024
    ) -> None:
        self.garbage_table = lgdo.Table(garbage_length)
        shape_guess = (garbage_length, packet_size_guess)
        self.garbage_table.add_field(
            "packets", lgdo.VectorOfVectors(shape_guess=shape_guess, dtype="uint8")
        )
        self.garbage_table.add_field(
            "packet_id", lgdo.Array(shape=garbage_length, dtype="uint32")
        )
        # TODO: add garbage codes enum attribute: user supplies in constructor
        # before calling super()
        self.garbage_table.add_field(
            "garbage_code", lgdo.Array(shape=garbage_length, dtype="uint32")
        )

    def get_key_lists(self) -> list[list[int | str]]:
        """Return a list of lists of keys available for this decoder.
        Each list must contain keys that can share a buffer, i.e. decoded_values
        is exactly the same (including e.g. waveform length) for all keys in the list.
        Overload with lists of keys for this decoder, e.g. ``return
        [range(n_channels)]``.  The default version works for decoders with
        single / no keys."""
        return [[None]]

    def get_decoded_values(self, key: int | str = None) -> dict:
        """Get decoded values (optionally for a given key, typically a channel).

        Notes
        -----
        Must overload for your decoder if it has key-specific decoded values.
        Must also implement ``key = None`` returns a "default"
        `decoded_values`. Otherwise, just returns :attr:`self.decoded_values`,
        which should be defined in the constructor.
        """
        if key is None:
            return self.decoded_values if hasattr(self, "decoded_values") else None

        raise NotImplementedError(
            "you need to implement key-specific get_decoded_values for",
            type(self).__name__,
        )

    def make_lgdo(self, key: int | str = None, size: int = None) -> LGDO:
        """Make an LGDO for this :class:`DataDecoder` to fill.

        This default version of this function allocates a
        :class:`~.lgdo.table.Table` using the `decoded_values` for key. If a
        different type of LGDO object is required for this decoder, overload
        this function.

        Parameters
        ----------
        key
            used to initialize the LGDO for a particular `key` (e.g.  to have
            different trace lengths for different channels of a piece of
            hardware). Leave as ``None`` if such specialization is not
            necessary.
        size
            the size to be allocated for the LGDO, if applicable.

        Returns
        -------
        data_obj
            the newly allocated LGDO.
        """

        if not hasattr(self, "decoded_values"):
            raise AttributeError(
                type(self).__name__
                + ": no decoded_values available for setting up table"
            )

        data_obj = lgdo.Table(size=size)
        dec_vals = self.get_decoded_values(key)
        for field, fld_attrs in dec_vals.items():
            # make a copy of fld_attrs: pop off the ones we use, then keep any
            # remaining user-set attrs and store into the lgdo
            attrs = fld_attrs.copy()

            # get the dtype
            if "dtype" not in attrs:
                raise AttributeError(
                    type(self).__name__ + ": must specify dtype for", field
                )

            dtype = attrs.pop("dtype")

            # no datatype: just a "normal" array
            if "datatype" not in attrs:
                # allow to override "kind" for the dtype for lgdo
                if "kind" in attrs:
                    attrs["datatype"] = "array<1>{" + attrs.pop("kind") + "}"
                data_obj.add_field(
                    field, lgdo.Array(shape=size, dtype=dtype, attrs=attrs)
                )
                continue

            # get datatype for complex objects
            datatype = attrs.pop("datatype")

            # waveforms: must have attributes t0_units, dt, dt_units, wf_len
            if datatype == "waveform":
                t0_units = attrs.pop("t0_units")
                dt = attrs.pop("dt")
                dt_units = attrs.pop("dt_units")
                wf_len = attrs.pop("wf_len")
                compression = attrs.pop("compression", None)

                wf_table = lgdo.WaveformTable(
                    size=size,
                    t0=0,
                    t0_units=t0_units,
                    dt=dt,
                    dt_units=dt_units,
                    wf_len=wf_len,
                    dtype=dtype,
                    attrs=attrs,
                )
                if compression is not None:
                    if not isinstance(compression, dict):
                        raise RuntimeError(
                            "waveform/compression attribute must be a dictionary"
                        )

                    if "values" in compression:
                        wf_table.values.attrs["compression"] = compression["values"]
                    if "t0" in compression:
                        wf_table.t0.attrs["compression"] = compression["t0"]
                    if "dt" in compression:
                        wf_table.dt.attrs["compression"] = compression["dt"]

                data_obj.add_field(field, wf_table)
                continue

            # Parse datatype for remaining lgdos
            datatype, shape, elements = lgdo.lgdo_utils.parse_datatype(datatype)

            # ArrayOfEqualSizedArrays
            if datatype == "array_of_equalsized_arrays":
                length = attrs.pop("length")
                # only arrays of 1D arrays are supported at present
                dims = (1, 1)
                aoesa = lgdo.ArrayOfEqualSizedArrays(
                    shape=(size, length), dtype=dtype, dims=dims, attrs=attrs
                )
                data_obj.add_field(field, aoesa)
                continue

            # VectorOfVectors
            if elements.startswith("array"):
                length_guess = size
                if "length_guess" in attrs:
                    length_guess = attrs.pop("length_guess")
                vov = lgdo.VectorOfVectors(
                    shape_guess=(size, length_guess), dtype=dtype, attrs=attrs
                )
                data_obj.add_field(field, vov)
                continue

            # if we get here, got a bad datatype
            raise RuntimeError(
                type(self).__name__,
                ": do not know how to make a",
                datatype,
                "for",
                field,
            )

        return data_obj

    def put_in_garbage(self, packet: int, packet_id: int, code: int) -> None:
        i_row = self.garbage_table.loc
        p8 = np.frombuffer(packet, dtype="uint8")
        self.garbage_table["packets"]._set_vector_unsafe(i_row, p8)
        self.garbage_table["packet_id"].nda[i_row] = packet_id
        self.garbage_table["garbage_codes"].nda[i_row] = code
        self.garbage_table.push_row()

    def write_out_garbage(
        self, filename: str, group: str = "/", lh5_store: LH5Store = None
    ) -> None:
        if lh5_store is None:
            lh5_store = lgdo.LH5Store()
        n_rows = self.garbage_table.loc
        if n_rows == 0:
            return
        lh5_store.write_object(
            self.garbage_table, "garbage", filename, group, n_rows=n_rows, append=True
        )
        self.garbage_table.clear()

    def get_max_rows_in_packet(self) -> int:
        """Returns the maximum number of rows that could be read out in a
        packet.

        1 by default, overload as necessary to avoid writing past the ends of
        buffers.
        """
        return 1

    def buffer_is_full(self, rb: RawBuffer) -> bool:
        """Returns whether the buffer is too full to read in another packet."""

        return len(rb.lgdo) - rb.loc < self.get_max_rows_in_packet()



"""
pygama's command line interface utilities.
"""
import argparse
import logging
import os
import sys

import numpy as np

import pygama
import pygama.logging
from pygama.dsp import build_dsp
from pygama.hit import build_hit
from pygama.lgdo import show
from pygama.raw import build_raw


def pygama_cli():
    """pygama's command line interface.

    Defines the command line interface (CLI) of the package, which exposes some
    of the most used functions to the console.  This function is added to the
    ``entry_points.console_scripts`` list and defines the ``pygama`` executable
    (see ``setuptools``' documentation). To learn more about the CLI, have a
    look at the help section:

    .. code-block:: console

      $ pygama --help
      $ pygama build-raw --help  # help section for a specific sub-command
    """

    parser = argparse.ArgumentParser(
        prog="pygama", description="pygama's command-line interface"
    )

    # global options
    parser.add_argument(
        "--version", action="store_true", help="""Print pygama version and exit"""
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="""Increase the program verbosity to maximum""",
    )

    subparsers = parser.add_subparsers()

    add_lh5ls_parser(subparsers)
    add_build_raw_parser(subparsers)
    add_build_dsp_parser(subparsers)
    add_build_hit_parser(subparsers)

    if len(sys.argv) < 2:
        parser.print_usage(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    if args.verbose:
        pygama.logging.setup(logging.DEBUG)
    elif args.debug:
        pygama.logging.setup(logging.DEBUG, logging.root)
    else:
        pygama.logging.setup()

    if args.version:
        print(pygama.__version__)  # noqa: T201
        sys.exit()

    args.func(args)


def add_lh5ls_parser(subparsers):
    """Configure :func:`.lgdo.lh5_store.show` command line interface."""

    parser_lh5ls = subparsers.add_parser(
        "lh5ls", description="""Inspect LEGEND HDF5 (LH5) file contents"""
    )
    parser_lh5ls.add_argument(
        "lh5_file",
        help="""Input LH5 file.""",
    )
    parser_lh5ls.add_argument(
        "lh5_group", nargs="?", help="""LH5 group.""", default="/"
    )
    parser_lh5ls.add_argument(
        "--attributes", "-a", action="store_true", help="""Print HDF5 attributes too"""
    )
    parser_lh5ls.set_defaults(func=lh5_show_cli)


def lh5_show_cli(args):
    """Passes command line arguments to :func:`.lgdo.lh5_store.show`."""

    show(args.lh5_file, args.lh5_group, attrs=args.attributes)


def add_build_raw_parser(subparsers):
    """Configure :func:`.raw.build_raw.build_raw` command line interface"""

    parser_d2r = subparsers.add_parser(
        "build-raw", description="""Convert data into LEGEND HDF5 (LH5) raw format"""
    )
    parser_d2r.add_argument(
        "in_stream",
        nargs="+",
        help="""Input stream. Can be a single file, a list of files or any
                other input type supported by the selected streamer""",
    )
    parser_d2r.add_argument(
        "--stream-type",
        "-t",
        help="""Input stream type name. Use this if the stream type cannot be
                automatically deduced by pygama""",
    )
    parser_d2r.add_argument(
        "--out-spec",
        "-o",
        help="""Specification for the output stream. HDF5 or JSON file name""",
    )
    parser_d2r.add_argument(
        "--buffer_size", "-b", type=int, default=8192, help="""Set buffer size"""
    )
    parser_d2r.add_argument(
        "--max-rows",
        "-n",
        type=int,
        default=np.inf,
        help="""Maximum number of rows of data to process from the input
                file""",
    )
    parser_d2r.add_argument(
        "--overwrite", "-w", action="store_true", help="""Overwrite output files"""
    )

    parser_d2r.set_defaults(func=build_raw_cli)


def build_raw_cli(args):
    """Passes command line arguments to :func:`.raw.build_raw.build_raw`."""

    for stream in args.in_stream:
        basename = os.path.splitext(os.path.basename(stream))[0]
        build_raw(
            stream,
            in_stream_type=args.stream_type,
            out_spec=args.out_spec,
            buffer_size=args.buffer_size,
            n_max=args.max_rows,
            overwrite=args.overwrite,
            orig_basename=basename,
        )


def add_build_dsp_parser(subparsers):
    """Configure :func:`.dsp.build_dsp.build_dsp` command line interface"""

    parser_r2d = subparsers.add_parser(
        "build-dsp",
        description="""Process LH5 raw files and produce a
        dsp file using a JSON configuration""",
    )
    parser_r2d.add_argument(
        "raw_lh5_file",
        nargs="+",
        help="""Input raw LH5 file. Can be a single file or a list of them""",
    )
    parser_r2d.add_argument(
        "--config",
        "-c",
        required=True,
        help=""""JSON file holding configuration of signal processing
                 routines""",
    )
    parser_r2d.add_argument(
        "--hdf5-groups",
        "-g",
        nargs="*",
        default=None,
        help="""Name of group in the LH5 file. By default process all base
                groups. Supports wildcards""",
    )
    parser_r2d.add_argument(
        "--output",
        "-o",
        default=None,
        help="""Name of output file, if only one is supplied. By default,
                output to <input-filename>_dsp.lh5""",
    )
    parser_r2d.add_argument(
        "--database",
        "-d",
        default=None,
        help="""JSON file to read database parameters from.  Should be nested
                dict with channel at the top level, and parameters below that""",
    )
    parser_r2d.add_argument(
        "--output-pars",
        "-p",
        nargs="*",
        default=None,
        help="""List of additional output DSP parameters written to file. By
                default use the "outputs" list defined in in the JSON
                configuration file""",
    )
    parser_r2d.add_argument(
        "--max-rows",
        "-n",
        default=None,
        type=int,
        help="""Number of rows to process. By default do the whole file""",
    )
    parser_r2d.add_argument(
        "--block",
        "-b",
        default=16,
        type=int,
        help="""Number of waveforms to process simultaneously. Default is
                16""",
    )
    parser_r2d.add_argument(
        "--chunk",
        "-k",
        default=3200,
        type=int,
        help="""Number of waveforms to read from disk at a time. Default is
                3200""",
    )

    group = parser_r2d.add_mutually_exclusive_group()
    group.add_argument(
        "--overwrite",
        "-w",
        action="store_const",
        const="r",
        dest="writemode",
        default="r",
        help="""Overwrite file if it already exists. Default option""",
    )
    group.add_argument(
        "--update",
        "-u",
        action="store_const",
        const="u",
        dest="writemode",
        help="""Update existing file with new values. Useful with the --output-pars
                option""",
    )
    group.add_argument(
        "--append",
        "-a",
        action="store_const",
        const="a",
        dest="writemode",
        help="""Append values to existing file""",
    )

    parser_r2d.set_defaults(func=build_dsp_cli)


def build_dsp_cli(args):
    """Passes command line arguments to :func:`.dsp.build_dsp.build_dsp`."""

    if len(args.raw_lh5_file) > 1 and args.output is not None:
        raise NotImplementedError("not possible to set multiple output file names yet")

    out_files = []
    if len(args.raw_lh5_file) == 1:
        if args.output is None:
            basename = os.path.splitext(os.path.basename(args.raw_lh5_file[0]))[0]
            basename = basename.removesuffix("_raw")
            out_files.append(f"{basename}_dsp.lh5")
        else:
            out_files.append(args.output)
    else:
        for file in args.raw_lh5_file:
            basename = os.path.splitext(os.path.basename(file))[0]
            basename = basename.removesuffix("_raw")
            out_files.append(f"{basename}_dsp.lh5")

    for i in range(len(args.raw_lh5_file)):
        build_dsp(
            args.raw_lh5_file[i],
            out_files[i],
            args.config,
            lh5_tables=args.hdf5_groups,
            database=args.database,
            outputs=args.output_pars,
            n_max=args.max_rows,
            write_mode=args.writemode,
            buffer_len=args.chunk,
            block_width=args.block,
        )


def add_build_hit_parser(subparsers):
    """Configure :func:`.hit.build_hit.build_hit` command line interface"""

    parser_r2d = subparsers.add_parser(
        "build-hit",
        description="""Process LH5 dsp files and produce a hit file using a
                       JSON configuration""",
    )
    parser_r2d.add_argument(
        "dsp_lh5_file",
        nargs="+",
        help="""Input dsp LH5 file. Can be a single file or a list of them""",
    )
    parser_r2d.add_argument(
        "--config",
        "-c",
        required=True,
        help=""""JSON file holding configuration of column operations on DSP output""",
    )
    parser_r2d.add_argument(
        "--hdf5-groups",
        "-g",
        nargs="*",
        default=None,
        help="""Name of group in the LH5 file. By default process all base
                groups. Supports wildcards""",
    )
    parser_r2d.add_argument(
        "--output",
        "-o",
        default=None,
        help="""Name of output file, if only one is supplied. By default,
                output to <input-filename>_hit.lh5""",
    )
    parser_r2d.add_argument(
        "--max-rows",
        "-n",
        default=None,
        type=int,
        help="""Number of rows to process. By default do the whole file""",
    )
    parser_r2d.add_argument(
        "--chunk",
        "-k",
        default=3200,
        type=int,
        help="""Number of waveforms to read from disk at a time. Default is
                3200""",
    )

    group = parser_r2d.add_mutually_exclusive_group()
    group.add_argument(
        "--overwrite",
        "-w",
        action="store_const",
        const="of",
        dest="writemode",
        default="w",
        help="""Overwrite file if it already exists. Default option""",
    )
    group.add_argument(
        "--update",
        "-u",
        action="store_const",
        const="u",
        dest="writemode",
        help="""Update existing file with new values""",
    )
    group.add_argument(
        "--append",
        "-a",
        action="store_const",
        const="a",
        dest="writemode",
        help="""Append values to existing file""",
    )

    parser_r2d.set_defaults(func=build_hit_cli)


def build_hit_cli(args):
    """Passes command line arguments to :func:`.hit.build_hit.build_hit`."""

    if len(args.dsp_lh5_file) > 1 and args.output is not None:
        raise NotImplementedError("not possible to set multiple output file names yet")

    out_files = []
    if len(args.dsp_lh5_file) == 1:
        if args.output is None:
            basename = os.path.splitext(os.path.basename(args.dsp_lh5_file[0]))[0]
            basename = basename.removesuffix("_dsp")
            out_files.append(f"{basename}_hit.lh5")
        else:
            out_files.append(args.output)
    else:
        for file in args.dsp_lh5_file:
            basename = os.path.splitext(os.path.basename(file))[0]
            basename = basename.removesuffix("_dsp")
            out_files.append(f"{basename}_hit.lh5")

    for i in range(len(args.dsp_lh5_file)):
        build_hit(
            args.dsp_lh5_file[i],
            args.config,
            outfile=out_files[i],
            lh5_tables=args.hdf5_groups,
            n_max=args.max_rows,
            wo_mode=args.writemode,
            buffer_len=args.chunk,
        )



from __future__ import annotations

import copy
import logging
from typing import Any

import fcutils

from pygama import lgdo
from pygama.raw.data_decoder import DataDecoder

log = logging.getLogger(__name__)

# put decoded values here where they can be used also by the orca decoder
fc_decoded_values = {
    # packet index in file
    "packet_id": {"dtype": "uint32"},
    # index of event
    "eventnumber": {"dtype": "int32"},
    # time since epoch
    "timestamp": {"dtype": "float64", "units": "s"},
    # time since beginning of file
    "runtime": {"dtype": "float64", "units": "s"},
    # number of triggered adc channels
    "numtraces": {"dtype": "int32"},
    # list of triggered adc channels
    "tracelist": {
        "dtype": "int16",
        "datatype": "array<1>{array<1>{real}}",  # vector of vectors
        "length_guess": 16,
    },
    # fpga baseline
    "baseline": {"dtype": "uint16"},
    # fpga energy
    "daqenergy": {"dtype": "uint16"},
    # right now, index of the trigger (trace)
    "channel": {"dtype": "uint32"},
    # PPS timestamp in sec
    "ts_pps": {"dtype": "int32"},
    # clock ticks
    "ts_ticks": {"dtype": "int32"},
    # max clock ticks
    "ts_maxticks": {"dtype": "int32"},
    # the offset in sec between the master and unix
    "mu_offset_sec": {"dtype": "int32"},
    # the offset in usec between master and unix
    "mu_offset_usec": {"dtype": "int32"},
    # the calculated sec which must be added to the master
    "to_master_sec": {"dtype": "int32"},
    # the delta time between master and unix in usec
    "delta_mu_usec": {"dtype": "int32"},
    # the abs(time) between master and unix in usec
    "abs_delta_mu_usec": {"dtype": "int32"},
    # startsec
    "to_start_sec": {"dtype": "int32"},
    # startusec
    "to_start_usec": {"dtype": "int32"},
    # start pps of the next dead window
    "dr_start_pps": {"dtype": "int32"},
    # start ticks of the next dead window
    "dr_start_ticks": {"dtype": "int32"},
    # stop pps of the next dead window
    "dr_stop_pps": {"dtype": "int32"},
    # stop ticks of the next dead window
    "dr_stop_ticks": {"dtype": "int32"},
    # maxticks of the dead window
    "dr_maxticks": {"dtype": "int32"},
    # current dead time calculated from deadregion (dr) fields.
    # Give the total dead time if summed up.
    "deadtime": {"dtype": "float64"},
    # waveform data
    "waveform": {
        "dtype": "uint16",
        "datatype": "waveform",
        "wf_len": 65532,  # max value. override this before initializing buffers to save RAM
        "dt": 16,  # override if a different clock rate is used
        "dt_units": "ns",
        "t0_units": "ns",
    },
}
"""Default FlashCam Event decoded values.

Re-used by :class:`~.raw.orca.orca_flashcam.ORFlashCamWaveformDecoder`.

Warning
-------
This configuration can be dynamically modified by the decoder at runtime.
"""


class FCEventDecoder(DataDecoder):
    """Decode FlashCam digitizer event data."""

    def __init__(self, *args, **kwargs) -> None:
        # these are read for every event (decode_event)
        self.decoded_values = copy.deepcopy(fc_decoded_values)
        super().__init__(*args, **kwargs)
        self.skipped_channels = {}
        self.fc_config = None
        self.max_numtraces = 1

    def get_key_lists(self) -> range:
        return [range(self.fc_config["nadcs"].value)]

    def get_decoded_values(self, channel: int = None) -> dict[str, dict[str, Any]]:
        # FC uses the same values for all channels
        return self.decoded_values

    def set_file_config(self, fc_config: lgdo.Struct) -> None:
        """Access ``FCIOConfig`` members once when each file is opened.

        Parameters
        ----------
        fc_config
            extracted via :meth:`~.fc_config_decoder.FCConfigDecoder.decode_config`.
        """
        self.fc_config = fc_config
        self.decoded_values["waveform"]["wf_len"] = self.fc_config["nsamples"].value

    def decode_packet(
        self,
        fcio: fcutils.fcio,
        evt_rbkd: lgdo.Table | dict[int, lgdo.Table],
        packet_id: int,
    ) -> bool:
        """Access ``FCIOEvent`` members for each event in the DAQ file.

        Parameters
        ----------
        fcio
            The interface to the ``fcio`` data. Enters this function after a
            call to ``fcio.get_record()`` so that data for `packet_id` ready to
            be read out.
        evt_rbkd
            A single table for reading out all data, or a dictionary of tables
            keyed by channel number.
        packet_id
            The index of the packet in the `fcio` stream. Incremented by
            :class:`~.raw.fc.fc_streamer.FCStreamer`.

        Returns
        -------
        n_bytes
            (estimated) number of bytes in the packet that was just decoded.
        """
        if fcio.numtraces > self.max_numtraces:
            self.max_numtraces = fcio.numtraces
            # The buffer might be storing all channels' data, so set the
            # fill_safety to the max number of traces we've seen so far.
            for rb in evt_rbkd.values():
                rb.fill_safety = self.max_numtraces
        any_full = False

        # a list of channels is read out simultaneously for each event
        for iwf in fcio.tracelist:
            if iwf not in evt_rbkd:
                if iwf not in self.skipped_channels:
                    # TODO: should this be a warning instead?
                    log.debug(f"skipping packets from channel {iwf}...")
                    self.skipped_channels[iwf] = 0
                self.skipped_channels[iwf] += 1
                continue
            tbl = evt_rbkd[iwf].lgdo
            if fcio.nsamples != tbl["waveform"]["values"].nda.shape[1]:
                log.warning(
                    "event wf length was",
                    fcio.nsamples,
                    "when",
                    self.decoded_values["waveform"]["wf_len"],
                    "were expected",
                )
            ii = evt_rbkd[iwf].loc

            # fill the table
            tbl["channel"].nda[ii] = iwf
            tbl["packet_id"].nda[ii] = packet_id
            tbl["eventnumber"].nda[
                ii
            ] = fcio.eventnumber  # the eventnumber since the beginning of the file
            tbl["timestamp"].nda[ii] = fcio.eventtime  # the time since epoch in seconds
            tbl["runtime"].nda[
                ii
            ] = fcio.runtime  # the time since the beginning of the file in seconds
            tbl["numtraces"].nda[ii] = fcio.numtraces  # number of triggered adcs
            tbl["tracelist"]._set_vector_unsafe(
                ii, fcio.tracelist
            )  # list of triggered adcs
            tbl["baseline"].nda[ii] = fcio.baseline[
                iwf
            ]  # the fpga baseline values for each channel in LSB
            tbl["daqenergy"].nda[ii] = fcio.daqenergy[
                iwf
            ]  # the fpga energy values for each channel in LSB
            tbl["ts_pps"].nda[ii] = fcio.timestamp_pps
            tbl["ts_ticks"].nda[ii] = fcio.timestamp_ticks
            tbl["ts_maxticks"].nda[ii] = fcio.timestamp_maxticks
            tbl["mu_offset_sec"].nda[ii] = fcio.timeoffset_mu_sec
            tbl["mu_offset_usec"].nda[ii] = fcio.timeoffset_mu_usec
            tbl["to_master_sec"].nda[ii] = fcio.timeoffset_master_sec
            tbl["delta_mu_usec"].nda[ii] = fcio.timeoffset_dt_mu_usec
            tbl["abs_delta_mu_usec"].nda[ii] = fcio.timeoffset_abs_mu_usec
            tbl["to_start_sec"].nda[ii] = fcio.timeoffset_start_sec
            tbl["to_start_usec"].nda[ii] = fcio.timeoffset_start_usec
            tbl["dr_start_pps"].nda[ii] = fcio.deadregion_start_pps
            tbl["dr_start_ticks"].nda[ii] = fcio.deadregion_start_ticks
            tbl["dr_stop_pps"].nda[ii] = fcio.deadregion_stop_pps
            tbl["dr_stop_ticks"].nda[ii] = fcio.deadregion_stop_ticks
            tbl["dr_maxticks"].nda[ii] = fcio.deadregion_maxticks
            tbl["deadtime"].nda[ii] = fcio.deadtime

            # if len(traces[iwf]) != fcio.nsamples: # number of sample per trace check
            tbl["waveform"]["values"].nda[ii][:] = fcio.traces[iwf]

            evt_rbkd[iwf].loc += 1
            any_full |= evt_rbkd[iwf].is_full()

        return any_full



import json
from pathlib import Path

import pytest
from pandas.testing import assert_frame_equal

from pygama.flow import FileDB

config_dir = Path(__file__).parent / "configs"


def test_init_partial(lgnd_test_data, test_filedb_full):
    with open(config_dir / "filedb-config.json") as f:
        config = json.load(f)

    config["data_dir"] = lgnd_test_data.get_path("lh5/prod-ref-l200/generated/tier")
    fdb = FileDB(config, scan=False)
    fdb.scan_files("cal/p01/r014")
    fdb.scan_tables_columns()

    assert fdb.df.equals(test_filedb_full.df)


def test_filedb_basics(test_filedb):
    db = test_filedb

    assert list(db.df.keys()) == [
        "exp",
        "period",
        "run",
        "timestamp",
        "type",
        "raw_file",
        "dsp_file",
        "hit_file",
        "tcm_file",
        "evt_file",
        "raw_size",
        "dsp_size",
        "hit_size",
        "tcm_size",
        "evt_size",
        "file_status",
    ]

    assert db.df.values.tolist() == [
        [
            "l60",
            "p01",
            "r014",
            "20220716T104550Z",
            "cal",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T104550Z-tier_raw.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T104550Z-tier_dsp.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T104550Z-tier_hit.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T104550Z-tier_tcm.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T104550Z-tier_evt.lh5",
            2141080,
            292032,
            280040,
            15728,
            0,
            int("0b11110", 2),
        ],
        [
            "l60",
            "p01",
            "r014",
            "20220716T105236Z",
            "cal",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T105236Z-tier_raw.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T105236Z-tier_dsp.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T105236Z-tier_hit.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T105236Z-tier_tcm.lh5",
            "/cal/p01/r014/l60-p01-r014-cal-20220716T105236Z-tier_evt.lh5",
            2141080,
            292032,
            280040,
            15728,
            0,
            int("0b11110", 2),
        ],
    ]


def test_scan_tables_columns(test_filedb_full):
    db = test_filedb_full

    assert list(db.df.keys()) == [
        "exp",
        "period",
        "run",
        "timestamp",
        "type",
        "raw_file",
        "dsp_file",
        "hit_file",
        "tcm_file",
        "evt_file",
        "raw_size",
        "dsp_size",
        "hit_size",
        "tcm_size",
        "evt_size",
        "file_status",
        "raw_tables",
        "raw_col_idx",
        "dsp_tables",
        "dsp_col_idx",
        "hit_tables",
        "hit_col_idx",
        "tcm_tables",
        "tcm_col_idx",
        "evt_tables",
        "evt_col_idx",
    ]

    assert db.columns == [
        [
            "baseline",
            "card",
            "ch_orca",
            "channel",
            "crate",
            "daqenergy",
            "deadtime",
            "dr_maxticks",
            "dr_start_pps",
            "dr_start_ticks",
            "dr_stop_pps",
            "dr_stop_ticks",
            "eventnumber",
            "fcid",
            "numtraces",
            "packet_id",
            "runtime",
            "timestamp",
            "to_abs_mu_usec",
            "to_dt_mu_usec",
            "to_master_sec",
            "to_mu_sec",
            "to_mu_usec",
            "to_start_sec",
            "to_start_usec",
            "tracelist",
            "ts_maxticks",
            "ts_pps",
            "ts_ticks",
            "waveform",
        ],
        [
            "bl_mean",
            "bl_std",
        ],
        ["hit_par1", "hit_par2"],
        ["array_id", "array_idx", "cumulative_length"],
    ]


def test_serialization(test_filedb_full):
    db = test_filedb_full
    db.to_disk("/tmp/filedb.lh5", wo_mode="of")

    with pytest.raises(RuntimeError):
        db.to_disk("/tmp/filedb.lh5")

    db2 = FileDB("/tmp/filedb.lh5")
    assert_frame_equal(db.df, db2.df)


def test_get_table_columns(test_filedb_full):
    db = test_filedb_full
    cols = db.get_table_columns(1, "dsp")
    assert cols == ["bl_mean", "bl_std"]
    with pytest.raises(KeyError):
        db.get_table_columns(1, "blah")
    with pytest.raises(ValueError):
        db.get_table_columns(9999, "raw")



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64[:])",
    ],
    "(n),(),()->(n)",
    **nb_kwargs,
)
def soft_pileup_corr(
    w_in: np.ndarray, n_in: int, tau_in: float, w_out: np.ndarray
) -> None:
    """Fit the baseline to an exponential with the provided time constant and
    then subtract the best-fit function from the entire waveform.

    Parameters
    ----------
    w_in
        The input waveform.
    n_in
        The number of samples at the beginning of the waveform
        to fit to an exponential.
    tau_in
        The fixed, exponential time constant.
    w_out
        The output waveform with the exponential subtracted.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_bl": {
            "function": "soft_pileup_corr",
            "module": "pygama.dsp.processors",
            "args": ["waveform", "1000", "500*us", "wf_bl"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(n_in) or np.isnan(tau_in):
        return

    if not np.floor(n_in) == n_in:
        raise DSPFatal("The number of samples is not an integer")

    if n_in < 2:
        raise DSPFatal("The number of samples is not enough for a fit")

    if n_in > len(w_in):
        raise DSPFatal("The number of samples is more than the waveform length")

    s1 = 0.0
    s2 = 0.0
    s3 = 0.0
    s4 = 0.0
    s5 = 0.0
    for i in range(0, n_in, 1):
        s1 += 1.0
        s2 += np.exp(-1.0 * i / tau_in)
        s3 += np.exp(-2.0 * i / tau_in)
        s4 += np.exp(-1.0 * i / tau_in) * w_in[i]
        s5 += w_in[i]
    b = (s5 - s2 * (s4 * s1 - s2 * s5) / (s3 * s1 - s2 * s2)) / s1
    a = (s4 - b * s2) / s3
    for i in range(0, len(w_in), 1):
        w_out[i] = w_in[i] - (a * np.exp(-1.0 * i / tau_in) + b)


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64[:])",
    ],
    "(n),(),(),()->(n)",
    **nb_kwargs,
)
def soft_pileup_corr_bl(
    w_in: np.ndarray, n_in: int, tau_in: float, b_in: float, w_out: np.ndarray
) -> None:
    """Fit the baseline to an exponential with the provided time
    constant and then subtract the best-fit function from the
    entire waveform.

    Parameters
    ----------
    w_in
        the input waveform.
    n_in
        the number of samples at the beginning of the waveform
        to fit to an exponential.
    tau_in
        the fixed, exponential time constant.
    b_in
        the fixed, exponential constant.
    w_out
        the output waveform with the exponential subtracted.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_bl": {
            "function": "soft_pileup_corr_bl",
            "module": "pygama.dsp.processors",
            "args": ["waveform", "1000", "500*us", "baseline", "wf_bl"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(n_in) or np.isnan(tau_in) or np.isnan(b_in):
        return

    if not np.floor(n_in) == n_in:
        raise DSPFatal("The number of samples is not an integer")

    if n_in < 1:
        raise DSPFatal("The number of samples is not enough for a fit")

    if n_in > len(w_in):
        raise DSPFatal("The number of samples is more than the waveform length")

    s2 = 0.0
    s3 = 0.0
    s4 = 0.0
    for i in range(0, n_in, 1):
        s2 += np.exp(-1.0 * i / tau_in)
        s3 += np.exp(-2.0 * i / tau_in)
        s4 += np.exp(-1.0 * i / tau_in) * w_in[i]
    a = (s4 - b_in * s2) / s3
    for i in range(0, len(w_in), 1):
        w_out[i] = w_in[i] - (a * np.exp(-1.0 * i / tau_in) + b_in)



from __future__ import annotations

import itertools
import math
import string
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas
import pint
from cycler import cycler
from matplotlib.lines import Line2D

import pygama.lgdo.lh5_store as lh5
from pygama.dsp.processing_chain import build_processing_chain
from pygama.math.units import unit_registry as ureg


class WaveformBrowser:
    """The :class:`WaveformBrowser` is a tool meant for interacting with
    waveforms from LEGEND HDF5 files. It defines an interface for drawing
    waveforms from a file, drawing transformed waveforms defined using
    :func:`~.dsp.build_dsp.build_dsp` style JSON files, drawing horizontal and
    vertical lines at the values of calculated parameters, and filling a legend
    with calculated parameters.
    """

    def __init__(
        self,
        files_in: str | list[str] | lgdo.LH5Iterator,  # noqa: F821
        lh5_group: str | list[str] = "",
        base_path: str = "",
        entry_list: list[int] | list[list[int]] = None,
        entry_mask: list[int] | list[list[int]] = None,
        dsp_config: str = None,
        database: str | dict = None,
        aux_values: pandas.DataFrame = None,
        lines: str | list[str] = "waveform",
        styles: dict[str, list] | str = None,
        legend: str | list[str] = None,
        legend_opts: dict = None,
        n_drawn: int = 1,
        x_unit: pint.Unit | str = None,
        x_lim: tuple[float | str | pint.Quantity] = None,
        y_lim: tuple[float | str | pint.Quantity] = None,
        norm: str = None,
        align: str = None,
        buffer_len: int = 128,
        block_width: int = 8,
    ) -> None:
        """
        Parameters
        ----------
        files_in
            name of file or list of files to browse. Can use wildcards. Can
            also pass an LH5Iterator

        lh5_group
            HDF5 group(s) to read. If a list is provided for both lh5_files
            and group, they must be the same size. If a file is wild-carded,
            the same group will be assigned to each file found
        base_path
            base path for file. See :class:`~.lgdo.lh5_store.LH5Store`.

        entry_list
            list of event indices to draw. If it is a nested list, use local
            indices for each file, otherwise use global indices.

        entry_mask
            boolean mask indicating which events to draw. If a nested list, use
            a mask for each file, else use a global mask. Cannot be used with
            `entry_list`.

        dsp_config
            name of DSP config JSON file containing a list of processors that
            can be applied to waveforms.

        database
            dictionary or JSON file with database of processing parameters.

        aux_values
            table of auxiliary values that are one-to-one with the input
            waveforms that can be drawn or placed in the legend.

        lines
            name(s) of objects to draw 2D lines for. Waveforms will be drawn as
            a time-series. Scalar quantities will be drawn as horizontal or
            vertical lines, depending on units. Vectors will be drawn as
            multiple horizontal or vertical lines.

        styles
            line colors and other style parameters to cycle through when
            drawing waveforms. Can be given as

            - dictionary of lists (e.g. ``{'color':['r', 'g', 'b'],
              'linestyle':['-', '--', '.']}``)
            - name of predefined style (see :mod:`matplotlib.style`
              documentation)
            - ``None`` (use current :obj:`matplotlib.rcParams` style).

            If a single style cycle is given, use for all lines; if a list is
            given, match to lines list.

        legend
            formatting string and values to include in the legend. This can
            be a list of values (one for each drawn object). If just a name
            is given, it will be auto-formatted to 3 digits. Otherwise,
            formatting strings in brackets can be used (e.g. ``{energy:0.1f}
            keV, {timestamp:d} ns``).  Names will be searched in the input
            file, DSP processed parameters, or auxiliary data-table.

        legend_opts
            dictionary containing additional keyword arguments for
            :mod:`matplotlib.legend`.

        n_drawn
            number of events to draw simultaneously when calling
            :meth:`draw_next`.

        x_lim, y_lim
            range of x- or y-values and units passed as tuple.

            - ``None``: Get range from first waveform drawn
            - :class:`pint.Quantity`: set value and x-unit
            - ``float``: get unit from first waveform drawn
            - ``str``: convert to :class:`pint.Quantity` (e.g. ``('0*us',
              '10*us')``).

        x_unit
            unit of x-axis.

        norm
            name of parameter (probably energy) to use to normalize waveforms.
            Useful when drawing multiple waveforms.

        align
            name of parameter to use for x-offset. Useful, e.g., for aligning
            multiple waveforms at a particular timepoint.

        buffer_len
            number of waveforms to keep in memory at a time.

        block_width
            block width for :class:`~.dsp.processing_chain.ProcessinChain`.
        """

        self.norm_par = norm
        self.align_par = align
        self.n_drawn = n_drawn
        self.next_entry = 0

        # data i/o initialization
        if isinstance(files_in, lh5.LH5Iterator):
            self.lh5_it = files_in
        else:
            # HACK: do not read VOV "tracelist", cannot be handled correctly by LH5Iterator
            # remove this hack once VOV support is properly implemented
            self.lh5_it = lh5.LH5Iterator(
                files_in,
                lh5_group,
                base_path=base_path,
                entry_list=entry_list,
                entry_mask=entry_mask,
                field_mask={"tracelist": False},
                buffer_len=buffer_len,
            )

        # Get the input buffer and read the first chunk
        self.lh5_in, _ = self.lh5_it.read(0)

        self.aux_vals = aux_values
        # Apply entry selection to aux_vals if needed
        if self.aux_vals is not None and len(self.aux_vals) > len(
            self.lh5_it.get_global_entrylist()
        ):
            self.aux_vals = self.aux_vals.iloc[
                self.lh5_it.get_global_entrylist()
            ].reset_index()

        # initialize objects to draw: dict from name to list of 2DLines
        if isinstance(lines, str):
            self.lines = {lines: []}
        elif lines is None:
            self.lines = {}
        else:
            self.lines = {line: [] for line in lines}

        # styles
        default_style = cycler(plt.rcParams["axes.prop_cycle"])
        if isinstance(styles, (list, tuple)):
            self.styles = [None for _ in self.lines]
            for i, sty in enumerate(styles):
                if isinstance(sty, str):
                    try:
                        self.styles[i] = plt.style.library[sty]["axes.prop_cycle"]
                    except KeyError:
                        self.styles[i] = itertools.repeat(None)
                elif sty is None:
                    self.styles[i] = default_style
                else:
                    self.styles[i] = cycler(**sty)
        else:
            if isinstance(styles, str):
                try:
                    self.styles = plt.style.library[styles]["axes.prop_cycle"]
                except KeyError:
                    self.styles = itertools.repeat(None)
            elif styles is None:
                self.styles = default_style
            else:
                self.styles = cycler(**styles)

        self.legend_format = []  # list of formatter strings
        self.legend_vals = {}  # Set up dict from names to lists of values

        if legend is None:
            legend = []
        elif isinstance(legend, str):
            legend = [legend]

        for entry in legend:
            legend_format = ""
            for st, name, form, cv in string.Formatter().parse(entry):
                if name is None:
                    # If name is none, this is the last batch of characters
                    legend_format += st
                    break

                if name == "":
                    raise KeyError("Cannot use empty formatter in " + entry)
                self.legend_vals[name] = []

                if form is None or form == "":
                    form = "~0.3P"
                cv = "" if cv is None or cv == "" else "!" + cv
                legend_format += f"{st}{{{name}:{form}{cv}}}"
            self.legend_format.append(legend_format)

        self.legend_kwargs = legend_opts if isinstance(legend_opts, dict) else {}

        # make processing chain and output buffer
        outputs = list(self.lines) + list(self.legend_vals)
        if isinstance(self.norm_par, str):
            outputs += [self.norm_par]
        if isinstance(self.align_par, str):
            outputs += [self.align_par]

        # Remove any values not found in aux_vals
        if self.aux_vals is not None:
            outputs = [o for o in outputs if o not in self.aux_vals]

        self.proc_chain, field_mask, self.lh5_out = build_processing_chain(
            self.lh5_in,
            dsp_config,
            db_dict=database,
            outputs=outputs,
            block_width=block_width,
        )
        self.lh5_it.reset_field_mask(field_mask)
        self.proc_chain.execute()

        # Check if all of our outputs can be found
        for name in outputs:
            if name not in self.lh5_out:
                raise KeyError(
                    "Could not find "
                    + name
                    + " in input lh5 file, DSP config file, or aux values"
                )

        self.x_unit = ureg(x_unit) if x_unit else None
        self.x_lim = x_lim
        self.y_lim = y_lim
        self.auto_x_lim = [np.inf, -np.inf]
        self.auto_y_lim = [np.inf, -np.inf]
        self.n_stored = 0

        # Set limit and convert to x-unit if needed; also set x_unit if needed
        if self.x_lim is not None:
            self.x_lim = list(self.x_lim)
            for i in range(2):
                if isinstance(self.x_lim[i], str):
                    self.x_lim[i] = ureg.Quantity(x_lim[i])
                if isinstance(self.x_lim[i], ureg.Quantity):
                    if self.x_unit:
                        self.x_lim[i] = float(self.x_lim[i] / self.x_unit)
                    else:
                        self.x_unit = self.x_lim[i].u
                        self.x_lim[i] = self.x_lim[i].m

        # If we still have no x_unit get it from the first waveform we can find
        if self.x_unit is None:
            for wf in self.lh5_out.values():
                if not isinstance(wf, lh5.WaveformTable):
                    continue
                self.x_unit = ureg(wf.dt_units)

        self.fig = None
        self.ax = None

    def new_figure(self) -> None:
        """Create a new figure and draw in it."""
        self.fig, self.ax = plt.subplots(1)

    def save_figure(self, f_out: str, *args, **kwargs) -> None:
        """Write figure to file named `f_out`. See
        :func:`matplotlib.pyplot.savefig` for `args` and `kwargs`."""
        self.fig.savefig(f_out)

    def set_figure(
        self, fig: WaveformBrowser | plt.Figure, ax: plt.Axes = None
    ) -> None:
        """Use an already existing figure and axis.

        Make sure to set ``clear=False`` when drawing if you don't want to
        clear what's already there! Can give a :class:`WaveformBrowser` object
        to use the figure / axis from that."""
        if isinstance(fig, WaveformBrowser):
            self.fig = fig.fig
            self.ax = fig.ax
        elif isinstance(fig, plt.Figure):
            self.fig = fig
            if ax is None:
                self.ax = fig.axes[0]
            elif isinstance(ax, plt.Axes):
                self.ax = ax
            else:
                raise TypeError("ax must be matplotlib.Axis")
        else:
            raise TypeError("fig must be matplotlib.Figure or WaveformBrowser")

    def clear_data(self) -> None:
        """Reset the currently stored data."""
        for line_data in self.lines.values():
            line_data.clear()
        for leg_data in self.legend_vals.values():
            leg_data.clear()
        self.auto_x_lim = [np.inf, -np.inf]
        self.auto_y_lim = [np.inf, -np.inf]
        self.n_stored = 0

    def find_entry(
        self, entry: int | list[int], append: bool = True, safe: bool = False
    ) -> None:
        """Find the requested data associated with entry in input files and
        place store it internally without drawing it.

        Parameters
        ----------
        entry
            index of entry or list of entries to find.
        append
            if ``False``, clear previously found data before finding more.
        safe
            if ``False``, throw an exception for out of range entries.
        """
        if not append:
            self.clear_data()
        if hasattr(entry, "__iter__"):
            for idx in entry:
                self.find_entry(idx)
            return

        # Get our current position in the I/O buffers; update if needed
        i_tb = entry - self.lh5_it.current_entry
        if not (self.lh5_it.n_rows > i_tb >= 0):
            self.lh5_it.read(entry)

            # Check if entry is out of range
            if self.lh5_it.n_rows == 0:
                if safe:
                    raise IndexError
                else:
                    return

            self.proc_chain.execute()
            i_tb = 0

        # get scaling factor/time shift if used
        if self.norm_par is None:
            norm = 1.0
        elif isinstance(self.norm_par, str):
            norm = self.lh5_out[self.norm_par].nda[entry]
        else:
            norm = self.norm_par[entry]

        if self.align_par is None:
            ref_time = 0
        elif isinstance(self.align_par, str):
            data = self.lh5_out.get(self.align_par, None)
            if isinstance(data, lh5.Array):
                ref_time = data.nda[i_tb]
                unit = data.attrs.get("units", None)
                if unit and unit in ureg and ureg.is_compatible_with(unit, self.x_unit):
                    ref_time *= float(ureg(unit) / self.x_unit)
            elif data is None:
                ref_time = self.aux_vals[self.align_par][entry]
            else:
                raise

        # lines
        lim = math.sqrt(sys.float_info.max)  # limits for v/h lines
        for name, lines in self.lines.items():
            # Get the data; note this is implicitly copying it!
            data = self.lh5_out.get(name, None)
            if isinstance(data, lh5.WaveformTable):
                y = data.values.nda[i_tb, :] / norm
                dt = data.dt.nda[i_tb] * float(ureg(data.dt_units) / self.x_unit)
                t0 = (
                    data.t0.nda[i_tb] * float(ureg(data.t0_units) / self.x_unit)
                    - ref_time
                )
                x = np.linspace(t0, t0 + dt * (data.wf_len - 1), data.wf_len)
                lines.append(Line2D(x, y))
                self._update_auto_limit(x, y)

            elif isinstance(data, lh5.ArrayOfEqualSizedArrays):
                y = data.nda[i_tb, :] / norm
                x = np.arange(len(y), dtype="float")
                lines.append(Line2D(x, y))
                self._update_auto_limit(x, y)

            elif isinstance(data, lh5.Array):
                val = data.nda[i_tb]
                unit = data.attrs.get("units", None)
                if unit and unit in ureg and ureg.is_compatible_with(unit, self.x_unit):
                    # Vertical line
                    val = np.array([val * float(ureg(unit) / self.x_unit) - ref_time])
                    lines.append(Line2D(np.tile(val, 2), [-lim, lim]))
                    self._update_auto_limit(val, None)
                else:
                    # Horizontal line
                    lines.append(Line2D([-lim, lim], np.tile(val / norm, 2)))
                    self._update_auto_limit(None, val)

            elif data is None:
                # Check for data in auxiliary table. It's unitless so I guess just do an hline...
                val = np.array([self.aux_vals[name][entry] / norm])
                lines.append(Line2D([-lim, lim], np.tile(val, 2)))
                self._update_auto_limit(None, val)

            else:
                raise TypeError(
                    f"Cannot draw '{name}'. WaveformBrowser does not support "
                    f"drawing lines for data of type {data.__class__}"
                )

        # legend data
        for name, vals in self.legend_vals.items():
            data = self.lh5_out.get(name, None)

            if not data:
                data = ureg.Quantity(self.aux_vals[name][entry])
            elif isinstance(data, lh5.Array):
                unit = data.attrs.get("units", None)
                if unit and unit in ureg:
                    data = data.nda[i_tb] * ureg(unit)
                else:
                    data = ureg.Quantity(data.nda[i_tb])
            else:
                raise TypeError(
                    "WaveformBrowser does not adding legend entries for data "
                    f"of type {data.__class__}"
                )

            vals.append(data)

        self.n_stored += 1
        self.next_entry = entry + 1

    def draw_current(self, clear: bool = True) -> None:
        """Draw the waveforms and data currently held internally by this class."""
        # Make figure/axis if needed
        if not (self.ax and self.fig and plt.fignum_exists(self.fig.number)):
            self.new_figure()

        if clear:
            self.ax.clear()

        x_lim = self.x_lim if self.x_lim else self.auto_x_lim
        y_lim = self.y_lim
        if not y_lim:
            y_range = self.auto_y_lim[1] - self.auto_y_lim[0]
            y_lim = [
                self.auto_y_lim[0] - 0.05 * y_range,
                self.auto_y_lim[1] + 0.05 * y_range,
            ]
        self.ax.set_xlim(*x_lim)
        self.ax.set_ylim(*y_lim)

        leg_handles = []
        leg_labels = []
        if not isinstance(self.styles, list):
            styles = self.styles

        # draw lines
        default_style = cycler(plt.rcParams["axes.prop_cycle"])
        for i, lines in enumerate(self.lines.values()):
            if isinstance(self.styles, list):
                styles = self.styles[i]
            if styles is None:
                styles = default_style

            for line, sty in zip(lines, itertools.cycle(styles)):
                if sty is not None:
                    line.update(sty)
                if line.get_figure() is not None:
                    line.remove()
                line.set_transform(self.ax.transData)
                self.ax.add_line(line)

                leg_handles.append(line)

        # Get legend entries
        try:
            leg_cycle = cycler(**self.legend_vals)
        except Exception:
            for form in self.legend_format:
                for _ in range(self.n_stored):
                    leg_labels.append(form)
        else:
            for form in self.legend_format:
                for leg_dat in leg_cycle:
                    leg_labels.append(form.format(**leg_dat))

        # Draw legend
        self.ax.set_xlabel(self.x_unit)
        if self.x_lim:
            self.ax.set_xlim(*self.x_lim)
        if len(leg_labels) > 0:
            if not clear:
                old_leg = self.ax.get_legend()
                if old_leg:
                    leg_handles = old_leg.get_lines() + leg_handles
                    leg_labels = [
                        t.get_text() for t in old_leg.get_texts()
                    ] + leg_labels
            self.ax.legend(leg_handles, leg_labels, **self.legend_kwargs)

    def _update_auto_limit(self, x: np.ndarray, y: np.ndarray) -> None:
        # Helper to update the automatic limits
        where = True

        if x is not None:
            where &= np.isfinite(x)
            if self.x_lim is not None:
                where &= (x >= self.x_lim[0]) & (x <= self.x_lim[1])

        if y is not None:
            where &= np.isfinite(y)
            if self.y_lim is not None:
                where &= (y >= self.y_lim[0]) & (y <= self.y_lim[1])

        if x is not None:
            self.auto_x_lim = [
                np.amin(x, where=where, initial=self.auto_x_lim[0]),
                np.amax(x, where=where, initial=self.auto_x_lim[1]),
            ]

        if y is not None:
            self.auto_y_lim = [
                np.amin(y, where=where, initial=self.auto_y_lim[0]),
                np.amax(y, where=where, initial=self.auto_y_lim[1]),
            ]

    def draw_entry(
        self,
        entry: int | list[int],
        append: bool = False,
        clear: bool = True,
        safe: bool = False,
    ) -> None:
        """Draw specified entry in the current figure/axes.

        Parameters
        ----------
        entry
            entry or list of entries to draw.
        append
            if ``True``, do not clear previously drawn entries before drawing more.
        clear
            if ``True``, clear previously drawn objects in the axes before drawing.
        safe
            if ``False``, throw an exception for out of range entries.
        """
        self.find_entry(entry, append)
        self.draw_current(clear)

    def find_next(self, n_wfs: int = None, append: bool = False) -> tuple[int, int]:
        """Find the next `n_wfs` waveforms (default `self.n_drawn`). See
        :meth:`find_entry`."""
        if not n_wfs:
            n_wfs = self.n_drawn
        entries = (self.next_entry, self.next_entry + n_wfs)
        self.find_entry(range(*entries), append, safe=True)
        return entries

    def draw_next(
        self, n_wfs: int = None, append: bool = False, clear: bool = True
    ) -> tuple[int, int]:
        """Draw the next `n_wfs` waveforms (default `self.n_drawn`). See
        :meth:`draw_next`."""
        entries = self.find_next(n_wfs, append)
        self.draw_current(clear)
        return entries

    def reset(self) -> None:
        """Reset to the start of the file for :meth:`draw_next`."""
        self.clear_data()
        self.next_entry = 0

    def __iter__(self) -> tuple[int, int]:
        while self.next_entry < len(self.lh5_it):
            yield self.draw_next()



from __future__ import annotations

from math import ceil, floor

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),(),(m)",
    **nb_kwargs,
)
def upsampler(w_in: np.ndarray, upsample: float, w_out: np.ndarray) -> None:
    """Upsamples the waveform by the number specified.

    Note
    ----
    A series of moving windows should be applied afterwards for smoothing.

    Parameters
    ----------
    w_in
        waveform to upsample.
    upsample
        number of samples to increase each sample to.
    w_out
        output array for upsampled waveform.
    """

    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    if not (upsample > 0):
        raise DSPFatal("Upsample must be greater than 0")

    for t_in in range(0, len(w_in)):
        t_out = int(t_in * upsample - np.floor(upsample / 2))
        for _ in range(0, int(upsample)):
            if (t_out >= 0) & (t_out < len(w_out)):
                w_out[t_out] = w_in[t_in]
            t_out += 1


@guvectorize(
    ["void(float32[:], char, float32[:])", "void(float64[:], char, float64[:])"],
    "(n),(),(m)",
    **nb_kwargs,
)
def interpolating_upsampler(
    w_in: np.ndarray, mode_in: np.int8, w_out: np.ndarray
) -> None:
    """Upsamples the waveform.

    Resampling ratio is set by size of `w_out` and `w_in`. Using the
    interpolation technique specified by `mode_in` to fill newly added samples.

    Parameters
    ----------
    w_in
        waveform to upsample.
    mode_in
        character selecting which interpolation method to use. Note this
        must be passed as a ``int8``, e.g. ``ord('i')``. Options:

        * ``i`` -- only set values at original samples; fill in between
          with zeros. Requires integer resampling ratio
        * ``n`` -- nearest-neighbor interpolation; defined at all values,
          but not continuous
        * ``f`` -- floor, or value at previous neighbor; defined at all
          values but not continuous
        * ``c`` -- ceiling, or value at next neighbor; defined at all values,
          but not continuous
        * ``l`` -- linear interpolation; continuous at all values, but not
          differentiable
        * ``h`` -- Hermite cubic spline interpolation; continuous and
          differentiable at all values but not twice-differentiable
        * ``s`` -- natural cubic spline interpolation; continuous and twice-
          differentiable at all values. This method is much slower
          than the others because it utilizes the entire input waveform!
    w_out
        output array for upsampled waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_up": {
            "function": "interpolating_upsampler",
            "module": "pygama.dsp.processors",
            "args": ["wf", "'s'", "wf_up(len(wf)*10, period=wf.period/10)"],
            "unit": "ADC"
        }
    """

    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    upsample = len(w_out) / len(w_in)

    if mode_in == ord("i"):  # Index
        if upsample != int(upsample):
            raise DSPFatal(
                "interpolating_upsampler requires len(w_out) to be an integer multiple of len(w_in) for mode 'i'"
            )
        for i_in, a in enumerate(w_in):
            i_out = int(upsample * i_in)
            w_out[i_out] = a
            w_out[i_out + 1 : i_out + upsample] = 0

    elif mode_in == ord("n"):  # Nearest-neighbor
        i_last = 0
        for i_in, a in enumerate(w_in):
            i_next = ceil(upsample * (i_in + 0.5))
            w_out[i_last:i_next] = a
            i_last = i_next
        w_out[i_last:] = w_in[-1]

    elif mode_in == ord("f"):  # Floor
        i_last = 0
        for i_in, a in enumerate(w_in):
            i_next = ceil(upsample * (i_in + 1))
            w_out[i_last:i_next] = a
            i_last = i_next

    elif mode_in == ord("c"):  # Ceiling
        i_last = 0
        for i_in, a in enumerate(w_in):
            i_next = floor(upsample * (i_in)) + 1
            w_out[i_last:i_next] = a
            i_last = i_next
        w_out[i_last:] = w_in[-1]

    elif mode_in == ord("l"):  # linear
        i_last = 0
        for i_in, a in enumerate(w_in):
            i_next = ceil(upsample * (i_in + 1))
            a_next = w_in[i_in + 1] if i_in < len(w_in) - 1 else w_in[-1]
            for i_out in range(i_last, i_next):
                t = i_out / upsample - i_in
                w_out[i_out] = a + t * (a_next - a)
            i_last = i_next

    elif mode_in == ord("h"):  # Cubic hermite
        i_last = 0
        for i_in in range(len(w_in) - 1):
            i_next = ceil(upsample * (i_in + 1))
            m0 = (
                w_in[1] - w_in[0]
                if i_in == 0
                else (w_in[i_in + 1] - w_in[i_in - 1]) / 2
            )
            m1 = (
                w_in[-1] - w_in[-2]
                if i_in == len(w_in) - 2
                else (w_in[i_in + 2] - w_in[i_in]) / 2
            )
            for i_out in range(i_last, i_next):
                t0 = i_out / upsample - i_in
                t1 = 1 - t0
                w_out[i_out] = (
                    (-2 * t1**3 + 3 * t1**2) * w_in[i_in]
                    + (-2 * t0**3 + 3 * t0**2) * w_in[i_in + 1]
                    - (t1**3 - t1**2) * m0
                    + (t0**3 - t0**2) * m1
                )
            i_last = i_next

        for i_out in range(i_last, len(w_out)):
            t0 = i_out / upsample - i_in
            t1 = 1 - t0
            w_out[i_out] = (
                (-2 * t1**3 + 3 * t1**2) * w_in[i_in]
                + (-2 * t0**3 + 3 * t0**2) * w_in[i_in + 1]
                - (t1**3 - t1**2) * m0
                + (t0**3 - t0**2) * m1
            )

    elif mode_in == ord("s"):  # Cubic spline
        u = np.zeros(len(w_in))
        w2 = np.zeros(len(w_in))  # second derivative values

        for i_in in range(1, len(w_in) - 1):
            p = 0.5 * w2[i_in - 1] + 2
            w2[i_in] = -0.5 / p
            u[i_in] = w_in[i_in + 1] - 2 * w_in[i_in] + w_in[i_in - 1]
            u[i_in] = (3 * u[i_in] - 0.5 * u[i_in - 1]) / p

        i_last = len(w_out)
        for i_in in range(len(w_in) - 2, -1, -1):
            w2[i_in] = w2[i_in] * w2[i_in + 1] + u[i_in]
            i_next = ceil(upsample * (i_in))

            for i_out in range(i_last, i_next - 1, -1):
                t0 = i_out / upsample - i_in
                t1 = 1 - t0
                w_out[i_out] = (
                    t1 * w_in[i_in]
                    + t0 * w_in[i_in + 1]
                    + ((t1**3 - t1) * w2[i_in] + (t0**3 - t0) * w2[i_in + 1]) / 6.0
                )
            i_last = i_next

    else:
        raise DSPFatal("Unrecognized interpolation mode")
